/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "icqtransport.h"

void it_message_url(session s, UIN_t uin, icqbyte *data, unsigned short len);
void it_message_contacts(session s, UIN_t uin, icqbyte *data, unsigned short len);
void it_message_pager(session s, unsigned short type, icqbyte *data, unsigned short len);
void it_message_auth_request(session s, UIN_t uin, icqbyte *data, unsigned short len);

static unsigned long it_extract_number(icqbyte **data,  unsigned short *len);
char *it_extract_string(pool p, icqbyte **data, unsigned short *len, int fix_crs);

void it_icq_message(session s, UIN_t uin, char *ts, unsigned short type, icqbyte *data, unsigned short len)
{
    xmlnode x;
    jid id;

    log_debug(ZONE,"message_incoming");

    /* normalize the incoming type */
    type &= (0x7FFF & ~(MSGTYPE_FLAG_MASS));

    switch (type)
    {
    case MSGTYPE_TEXT:
    {
        /* ordinary text message */
        x = xmlnode_new_tag("message");
        xmlnode_insert_cdata(xmlnode_insert_tag(x,"subject"),"ICQ Message",-1);
        xmlnode_insert_cdata(xmlnode_insert_tag(x,"body"),
                             it_extract_string(xmlnode_pool(x),&data,&len,1),-1);

        xmlnode_put_attrib(x,"to",jid_full(s->id));
        id = it_uin2jid(xmlnode_pool(x),uin,s->from->server);
        xmlnode_put_attrib(x,"from",jid_full(id));

        if (ts) it_delay(x,ts);

        it_deliver(s->ti,x);
        break;
    }

    case MSGTYPE_AUTHORIZE:
    {
        contact c = it_contact_get(s,uin);
        if (c == NULL)  /* XXX temp hack.  probably means they went offline,
                           came back on and then received this auth */
            c = it_contact_add(s,uin);

        if (c->status == STATUS_NOT_IN_LIST) /* no point in subscribing twice */
            mtq_send(s->q,c->p,it_contact_subscribe,(void *) c);
        break;
    }

    case MSGTYPE_AUTH_DENY:
    {
        contact c = it_contact_get(s,uin);
        if (c && c->subscribing)
        {
            pool p;

            c->subscribing = 0;
            x = jutil_presnew(JPACKET__UNSUBSCRIBED,jid_full(s->id),NULL);
            p = xmlnode_pool(x);
            id = it_uin2jid(p,uin,s->from->server);
            xmlnode_put_attrib(x,"from",jid_full(id));
            xmlnode_insert_cdata(xmlnode_insert_tag(x,"status"),it_extract_string(p,&data,&len,1),-1);
            it_deliver(s->ti,x);
        }
        else
            log_debug(ZONE,"Session[%s], received AUTH_DENY from uknown contact %lu",jid_full(s->id),uin);
        break;
    }

    case MSGTYPE_URL:
        it_message_url(s,uin,data,len);
        break;

    case MSGTYPE_CONTACT_ADD:   /* you were added to contact list */
        x = jutil_presnew(JPACKET__SUBSCRIBED,jid_full(s->id),"The user added you to their contact list.");
        xmlnode_put_attrib(x,"from",jid_full(it_uin2jid(xmlnode_pool(x),uin,s->from->server)));
        it_deliver(s->ti,x);
        break;

    case MSGTYPE_AUTH_REQUEST:  /* request for authorization */
        it_message_auth_request(s,uin,data,len);
        break;

    case MSGTYPE_WEB_PAGER:     /* Web pager message */
    case MSGTYPE_EMAIL_PAGER:   /* email pager message */
        it_message_pager(s,type,data,len);
        break;

    case MSGTYPE_CONTACTS:
        it_message_contacts(s,uin,data,len);
        break;

    case MSGTYPE_GREETING_CARD:
        break;

    default:
        log_debug(ZONE,"warning: got invalid message type %u",type);
        break;
    }
}

void it_message_url(session s, UIN_t uin, icqbyte *data, unsigned short len)
{
    pool p;
    spool sp;
    xmlnode msg, x;
    char *text, *data_value;

    /* URL message */
    msg = xmlnode_new_tag("message");
    p = xmlnode_pool(msg);
    text = it_extract_string(p,&data,&len,1);
    data_value = it_extract_string(p,&data,&len,0);

    xmlnode_insert_cdata(xmlnode_insert_tag(msg,"subject"),"ICQ URL Message",-1);

    /* add a message body that most clients can understand */
    sp = spool_new(p);
    spooler(sp,"URL: ",data_value,"\nDescription:\n",text,sp);

    xmlnode_insert_cdata(xmlnode_insert_tag(msg,"body"),spool_print(sp),-1);

    x = xmlnode_insert_tag(msg,"x");
    xmlnode_put_attrib(x,"xmlns",NS_XOOB);
    xmlnode_insert_cdata(xmlnode_insert_tag(x,"desc"),text,-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(x,"url") ,data_value,-1);

    xmlnode_put_attrib(msg,"to",jid_full(s->id));
    xmlnode_put_attrib(msg,"from",jid_full(it_uin2jid(p,uin,s->from->server)));

    it_deliver(s->ti,msg);
}

void it_message_contacts(session s, UIN_t from, icqbyte *data, unsigned short len)
{
    xmlnode x, q, item;
    pool p;
    spool sp;
    UIN_t uin;
    char *nick, *id;
    int i, nc;

    x = xmlnode_new_tag("message");
    p = xmlnode_pool(x);
    xmlnode_insert_cdata(xmlnode_insert_tag(x,"subject"),"Roster item(s)",-1);

    q = xmlnode_insert_tag(x,"x");
    xmlnode_put_attrib(q,"xmlns","jabber:x:roster");
    sp = spool_new(p);

    nc = it_extract_number(&data,&len);
    for (i = 0; i < nc; i++)
    {
        nick = (char *)it_extract_string(p,&data,&len,0);
        uin = it_extract_number(&data,&len);

        id = jid_full(it_uin2jid(p,uin,s->from->server));
        item = xmlnode_insert_tag(q,"item");
        xmlnode_put_attrib(item,"jid",id);
        xmlnode_put_attrib(item,"name",nick);

        spooler(sp,nick," <",id,">\n",sp);
    }

    xmlnode_insert_cdata(xmlnode_insert_tag(x,"body"),spool_print(sp),-1);

    xmlnode_put_attrib(x,"to",jid_full(s->id));
    xmlnode_put_attrib(x,"from",jid_full(it_uin2jid(p,uin,s->from->server)));

    it_deliver(s->ti,x);
}

/* skips a field header at the beginning of a line, and returns a pointer to the first data character on that line. */
char *it_skip_field_header(char *ptr_start, char *ptr_end)
{
    while ((ptr_start<ptr_end) && (*ptr_start!=':'))
        ptr_start++;  /* skip the name */
    if (ptr_start<ptr_end)
        ptr_start++;  /* skip the colon */
    while ((ptr_start<ptr_end) && isspace(*ptr_start))
        ptr_start++;  /* skip leading whitespace */

    return ptr_start;
}

void it_message_pager(session s, unsigned short type, icqbyte *data, unsigned short len)
{
    char *sip_start, *sip_end;    /* sender IP start & end */
    char *subj_start, *subj_end;  /* subject start and end */
	char *real_body;              /* REAL message body */
	char *sz_sip, *sz_subj;             /* copied strings */
    xmlnode msg;
    spool sp;
    char *nick, *first, *last, *email, *data_value, *text;
    int auth_flag;
    pool p;

    /* begin assembling the message */
	msg = xmlnode_new_tag("message");
    xmlnode_put_attrib(msg,"from",jid_full(s->from));
    xmlnode_put_attrib(msg,"to",jid_full(s->id));
    p = xmlnode_pool(msg);

    nick = it_extract_string(p,&data,&len,0);
    first = it_extract_string(p,&data,&len,0);
    last = it_extract_string(p,&data,&len,0);
    email = it_extract_string(p,&data,&len,0);
    auth_flag = data ? (unsigned char)(*data & 0x0F) : 0;

    data_value = it_extract_string(p,&data,&len,1);
    text = it_extract_string(p,&data,&len,1);
    if (text == NULL) text = data_value;
    data_value = NULL;

	if (type == MSGTYPE_WEB_PAGER)
	{
        /* figure out sender IP start and end, and subject start */
        sip_start = text;
        sip_end = strchr(sip_start,'\n');
        subj_start = sip_end + 1;
        sip_start = it_skip_field_header(sip_start,sip_end);
        if (sip_start!=sip_end)
        {
            /* copy out the sender IP string */
            sz_sip = (char *)alloca(((int)(sip_end-sip_start))+1);
            memcpy(sz_sip,sip_start,(int)(sip_end-sip_start));
            sz_sip[(int)(sip_end-sip_start)] = '\0';
        }
        else /* no sender IP */
            sz_sip = NULL;
	}
	else
	{
        /* easier to find subject start, as there IS no Sender IP */
        sip_start = sip_end = NULL;
        sz_sip = NULL;
        subj_start = text;
	}

	/* figure out subject end and real body start */
	subj_end = strchr(subj_start,'\n');
	real_body = subj_end + 1;
	subj_start = it_skip_field_header(subj_start,subj_end);
	if (subj_start != subj_end)
	{
        /* copy out the subject string */
        sz_subj = (char *)alloca(((int)(subj_end-subj_start))+1);
        memcpy(sz_subj,subj_start,(int)(subj_end-subj_start));
        sz_subj[(int)(subj_end-subj_start)] = '\0';
	}
	else
        sz_subj = NULL;

	/* build the subject */
	sp = spool_new(p);
	spooler(sp,"ICQ ",((type == MSGTYPE_WEB_PAGER) ? "Web" : "Email")," Pager Message",sp);
	if (sz_subj)
        spooler(sp,": ",sz_subj,sp);

	xmlnode_insert_cdata(xmlnode_insert_tag(msg,"subject"),spool_print(sp),-1);

	/* build the message body */
	sp = spool_new(p);
	spooler(sp,"Sender: ",nick," <",email,">",sp);

	if (sz_sip)
        spooler(sp," @[",sz_sip,"]",sp);

	spooler(sp,"\n\n",real_body,sp);

	xmlnode_insert_cdata(xmlnode_insert_tag(msg,"body"),spool_print(sp),-1);

    it_deliver(s->ti,msg);
}

void it_message_auth_request(session s, UIN_t uin, icqbyte *data, unsigned short len)
{
    contact c;
    xmlnode x;
    pool p;
    char *nick, *first, *last, *email, *data_value, *text;
    char *reason;
    int auth_flag;

    x = jutil_presnew(JPACKET__SUBSCRIBE,jid_full(s->id),NULL);
    p = xmlnode_pool(x);
    xmlnode_put_attrib(x,"from",jid_full(it_uin2jid(p,uin,s->from->server)));

    nick = it_extract_string(p,&data,&len,0);
    first = it_extract_string(p,&data,&len,0);
    last = it_extract_string(p,&data,&len,0);
    email = it_extract_string(p,&data,&len,0);
    auth_flag = data ? (unsigned char)(*data & 0x0F) : 0;

    data_value = it_extract_string(p,&data,&len,1);
    text = it_extract_string(p,&data,&len,1);
    if (text == NULL) text = data_value;
    data_value = NULL;

    reason = spools(p,nick," (",first,", ",last,") ",text,p);
    xmlnode_insert_cdata(xmlnode_insert_tag(x,"status"),reason,-1);

    c = it_contact_get(s,uin);
    if (c == NULL)
        c = it_contact_add(s,uin);

    c->asking = 1;

    it_deliver(s->ti,x);
}

/* extracts a (possibly-delimited) string from a message buffer, converts
   it to UTF-8 (from Windows-1252), and moves the data pointer and length to a point after
   the extracted string.  Returns the converted UTF-8 string. */
char *it_extract_string(pool p, icqbyte **data, unsigned short *len, int fix_crs)
{
    icqbyte *start = *data;  /* start pointer */
    icqbyte *ptr;            /* end pointer */
    const char *inter;       /* intermediate step for conversion */

    if (!data || !len || !(*data) || ((*len)==0))
        return NULL;  /* safety check */

    ptr = (icqbyte *)memchr(*data,SEPARATOR_CHAR,*len);  /* end pointer */
    if (ptr)
    {
        /* terminate the string and adjust to go to the next one */
        *ptr++ = '\0';
        *len -= (unsigned)(ptr - *data);
        *data = ptr;
        if (fix_crs)
            inter = (const char *)it_strrepl(p,(const char *)start,s_CRLF,s_LF);
        else
            inter = (const char *)start;
        return (char *) convert_windows_to_utf8(p,inter);   
    }
    else
    {
        /* look for a string terminator character */
        ptr = (icqbyte *)memchr(*data,'\0',*len);
        if (ptr)
        {
            /* the NULL-terminated string is the last one... */
            *len = 0;
            *data = NULL;
            if (fix_crs)
                inter = (const char *)it_strrepl(p,(const char *)start,s_CRLF,s_LF);
            else
                inter = (const char *)start;
            return (char *)convert_windows_to_utf8(p,inter);   
        }
        else
        {
            /* there is no terminator - just convert it raw */
            char *tmpbuf = (char *)alloca(*len + 1);

            memcpy(tmpbuf,start,*len);
            tmpbuf[*len] = '\0';
            *len = 0;
            *data = NULL;
            if (fix_crs)
                inter = (const char *)it_strrepl(p,(const char *)tmpbuf,s_CRLF,s_LF);
            else
                inter = (const char *)tmpbuf;
            return (char *)convert_windows_to_utf8(p,inter);
        }   
    }
}

/* extracts a (possibly-delimited) string from a message buffer, converts
   it to an unsigned long integer, and moves the data pointer and length to a point after
   the extracted string.  Returns the converted integer value. */
static unsigned long it_extract_number(icqbyte **data,  unsigned short *len)
{
    icqbyte *start = *data;  /* start pointer */
    icqbyte *ptr;            /* end pointer */

    if (!data || !len || !(*data) || ((*len)==0))
        return 0;  /* safety check */

    ptr = (icqbyte *)memchr(*data,SEPARATOR_CHAR,*len);  /* end pointer */
    if (ptr)
    { /* terminate the string and adjust to go to the next one */
        *ptr++ = '\0';
        *len -= (unsigned)(ptr - *data);
        *data = ptr;
        return strtoul((const char *)start,NULL,10);
    }
    else
    { /* look for a string terminator character */
        ptr = (icqbyte *)memchr(*data,'\0',*len);
        if (ptr)
        { /* the NULL-terminated string is the last one... */
            *len = 0;
            *data = NULL;
            return strtoul((const char *)start,NULL,10);
        }
        else
        { /* there is no terminator - just convert it raw */
            char *tmpbuf = (char *)alloca(*len + 1);
            memcpy(tmpbuf,start,*len);
            tmpbuf[*len] = '\0';
            *len = 0;
            *data = NULL;
            return strtoul(tmpbuf,NULL,10);   
        }   
    }
}
