/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "icqtransport.h"

void it_server_login(session s)
{
    iti ti = s->ti;
    ICQBUFFER5C packet;
    int len;
    unsigned offset;

    len = strlen(s->passwd) + 1;

    /* Assemble the packet data. */
    put_icqlong(packet.data,0,time(NULL));

    if (ti->disable_tcp == 0)
        put_icqlong(packet.data,4,ntohs(s->addr.sin_port));    /* the port number */
    else
        put_icqlong(packet.data,4,0);

    /* fill in password length and password. */
    put_icqshort(packet.data,8,(unsigned short) len);
    memcpy(packet.data + 10,s->passwd,len);
    offset = (unsigned) (len + 10);

    /* Fill in Part 2 of the login packet. */
    put_icqlong(packet.data,offset,LOGIN_MAGIC1);
    put_icqlong(packet.data,offset + 4,s->addr.sin_addr.s_addr);
    packet.data[offset + 8] = (icqbyte) ti->disable_tcp ? 0 :LOGIN_FLAG1_USETCP;
    put_icqlong(packet.data,offset + 9, s->status < 0 ? 0 : it_status_to_bits(s->status));
    put_icqshort(packet.data,offset + 13,TCP_VERSION_NUMBER);

    put_icqshort(packet.data,offset + 15,LOGIN_MAGIC2);
    put_icqlong(packet.data,offset + 17,LOGIN_MAGIC3);
    put_icqlong(packet.data,offset + 21,LOGIN_MAGIC4);
    put_icqlong(packet.data,offset + 25,LOGIN_MAGIC5);
    put_icqlong(packet.data,offset + 29,LOGIN_MAGIC6);
    put_icqlong(packet.data,offset + 33,LOGIN_MAGIC7);

    it_build_and_write(s,&packet,len + 47,CMD_LOGIN);
}

void it_server_login_1(session s)
{
    ICQBUFFER5C packet;

    put_icqlong(packet.data,0,random());
    it_build_and_write(s,&packet,4,CMD_LOGIN_1);
}

void it_server_new_UIN(session s)
{
    ICQBUFFER5C packet;
    int len;
    unsigned offset;

    len = strlen(s->passwd) + 1;
    if (len > (UDP_MAX_DATA - (2 + 16)))  /* password is to long? */
        return;

    /* fill in password length and password. */
    put_icqshort(packet.data,0,(unsigned short)len);
    memcpy(packet.data + 2,s->passwd,len);
    offset = (unsigned)(len + 2);

    /* Fill in Part 2 of the new register packet. */
    put_icqlong(packet.data,offset,REG_NEW_MAGIC1);
    put_icqlong(packet.data,offset+4,REG_NEW_MAGIC2);
    put_icqlong(packet.data,offset+8,REG_NEW_MAGIC3);
    put_icqlong(packet.data,offset+12,REG_NEW_MAGIC4);

    /* Prepare and send the packet. */
    it_build_and_write(s,&packet,offset+16,CMD_REG_NEW_USER);
}

void it_server_ack_messages(session s)
{
    ICQBUFFER5C packet;

    /* Fill in the random data and send it out. */
    put_icqlong(packet.data,0,random());
    it_build_and_write(s,&packet,4,CMD_ACK_MESSAGES);
}

void it_server_text_code(session s, const char *code)
{
    ICQBUFFER5C packet;
    int len;

    len = strlen(code) + 1;

    /* Fill in the text code information. */
    put_icqshort(packet.data,0,(unsigned short)len);
    memcpy(packet.data+2,code,len);
    put_icqshort(packet.data,(unsigned)(len+2),TEXT_CODE_MAGIC);

    /* Prepare and send the packet. */
    it_build_and_write(s,&packet,(unsigned)(len + 4),CMD_SEND_TEXT_CODE);
}

void it_server_keepalive(session s)
{
    ICQBUFFER5C packet;

    log_debug(ZONE,"Session[%s], keepalive",jid_full(s->id));

    s->timer = time(NULL);

    /* Fill in the random data and send it out. */
    put_icqlong(packet.data,0,random());
    it_build_and_write(s,&packet,4,CMD_KEEP_ALIVE);
}

void it_server_send_contact_list(session s)
{
    icqlong *uin_array;     /* pointer to UIN array within buffer */
    contact cur;
    ICQBUFFER5C buffer;     /* buffer for outgoing packet */
    int cct = 0;            /* contact count */
    int send_null = 1;      /* send a null at end? */

    log_debug(ZONE,"Session[%s], sending contact list",jid_full(s->id));

    uin_array = (icqlong *)(&(buffer.data[1]));  /* initialize helper pointer */

    for (cur = s->contacts; cur != NULL; cur = cur->next)
    {
        if (cur->status == STATUS_NOT_IN_LIST)
            continue;

        /* add contact to the UIN array */
        uin_array[cct++] = to_icqlong(cur->uin);

        if (cct == UDP_MAX_CONTACTS)
        {
            /* build and write a completed packet */
            buffer.data[0] = (icqbyte)UDP_MAX_CONTACTS;
            it_build_and_write(s,&buffer,(UDP_MAX_CONTACTS * 4) + 1,CMD_CONTACT_LIST);

            send_null = 0;
            cct = 0;
        }
    }

    if (cct > 0 || send_null)
    {
        buffer.data[0] = (icqbyte)cct;
        it_build_and_write(s,&buffer,(cct * 4) + 1,CMD_CONTACT_LIST);   
    }
}

void it_server_send_status(session s)
{
    ICQBUFFER5C packet;
    put_icqlong(packet.data,0,it_status_to_bits(s->status));
    it_build_and_write(s,&packet,4,CMD_STATUS_CHANGE);
}

void it_server_request_meta_info(session s, UIN_t uin)
{
    ICQBUFFER5C packet;

    put_icqshort(packet.data,0,META_GET_USER_INFO);
    put_icqlong(packet.data,2,uin);

    it_build_and_write(s,&packet,6,CMD_META_USER);
}

void it_server_send_info(session s)
{
    ICQBUFFER5C packet;
    int nick_len, first_len, last_len, email_len, len;
    unsigned offset = 0;

    nick_len = j_strlen(s->nick) + 1;
    first_len = j_strlen(s->first) + 1;
    last_len = j_strlen(s->last) + 1;
    email_len = j_strlen(s->email) + 1;
    len = nick_len + first_len + last_len + email_len + 4 + (2 * 4);

    if (len > (UDP_MAX_DATA - 15))
    {
        log_debug(ZONE,"to long");
        return;
    }

    put_icqshort(packet.data,offset,(unsigned short) nick_len);
    offset += 2;
    if (s->nick)
        memcpy(packet.data + offset,s->nick,nick_len);
    else
        packet.data[offset] = '\0';
    offset += nick_len;

    put_icqshort(packet.data,offset,(unsigned short) first_len);
    offset += 2;
    if (s->first)
        memcpy(packet.data + offset,s->first,first_len);
    else
        packet.data[offset] = '\0';
    offset += first_len;

    put_icqshort(packet.data,offset,(unsigned short) last_len);
    offset += 2;
    if (s->last)
        memcpy(packet.data + offset,s->last,last_len);
    else
        packet.data[offset] = '\0';
    offset += last_len;

    put_icqshort(packet.data,offset,(unsigned short) email_len);
    offset += 2;
    if (s->email)
        memcpy(packet.data + offset,s->email,email_len);
    else
        packet.data[offset] = '\0';
    offset += email_len;

    packet.data[offset + 1] = (icqbyte)0;
    packet.data[offset + 2] = (icqbyte)1;
    packet.data[offset + 3] = (icqbyte)1;
    packet.data[offset + 4] = (icqbyte)1;

    /* XXX not sure about this... */
    /* Set the "authorization required" flag on the new account. */
    // packet.data[0] = packet.data[1] = (icqbyte)0;

    it_build_and_write(s,&packet,len,CMD_AUTH_UPDATE);

    it_server_send_status(s);
}

void it_server_search_uin(session s, UIN_t uin)
{
    ICQBUFFER5C packet;
    int len;

    put_icqshort(packet.data,0,META_SEARCH_UIN);
    put_icqlong(packet.data,2,uin);
    len = 6;

    it_build_and_write(s,&packet,len,CMD_META_USER);
}

int it_server_search_email(session s, char *email)
{
    ICQBUFFER5C packet;
    int email_len, len;

    email_len = strlen(email) + 1;
    len = 4 + email_len;

    if (len > UDP_MAX_DATA)
        return 1;

    put_icqshort(packet.data,0,META_SEARCH_EMAIL);
    put_icqshort(packet.data,2,(unsigned short)email_len);
    memcpy(packet.data+4,email,email_len);

    /* Build and write the meta-search packet. */
    it_build_and_write(s,&packet,len,CMD_META_USER); 

    return 0;
}

int it_server_search_names(session s, char *nick, char *first, char *last)
{
    ICQBUFFER5C packet;
    int nick_len, first_len, last_len, len;
    unsigned offset;

    nick_len = strlen(nick) + 1;
    first_len = strlen(first) + 1;
    last_len = strlen(last) + 1;

    len = 8 + nick_len + first_len + last_len;
    if (len > UDP_MAX_DATA)
        return 1;

    put_icqshort(packet.data,0,META_SEARCH_NAME);

    put_icqshort(packet.data,2,(unsigned short) first_len);
    memcpy(packet.data + 4,first,first_len);
    offset = 4 + first_len;

    put_icqshort(packet.data,offset,(unsigned short) last_len);
    memcpy(packet.data+(offset + 2),last,last_len);
    offset += (2 + last_len);

    put_icqshort(packet.data,offset,(unsigned short) nick_len);
    memcpy(packet.data+(offset + 2),nick,nick_len);

    it_build_and_write(s,&packet,len,CMD_META_USER);

    return 0;
}

int it_server_send_message(session s, imessage m)
{
    ICQBUFFER5C packet;
    int len;

    len = it_message_size(m);

    /* fill in the packet data */
    put_icqlong(packet.data,0,m->to);
    put_icqshort(packet.data,4,m->type);
    put_icqshort(packet.data,6,(unsigned short)len);
    it_message_format(m,packet.data + 8);

    /* send out the packet */
    it_build_and_write(s,&packet,len + 8,CMD_SENDM);

    return 0;
}

int it_server_send_message_raw(session s, UIN_t uin, unsigned short type, unsigned len, icqbyte *data)
{
    ICQBUFFER5C packet;

    put_icqlong(packet.data,0,uin);
    put_icqshort(packet.data,4,type);
    put_icqshort(packet.data,6,(unsigned short)len);
    memcpy(packet.data+8,data,len);

    it_build_and_write(s,&packet,len + 8,CMD_SENDM);

    return 0;
}
