/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "jabberd.h"
#include "protocol.h"

#define RESEND_DELAY 10
#define MAX_RETRY_ATTEMPTS 8
#define KEEPALIVE 30
#define LOGIN_TIMER 60
#define RSEQ_MAX           64      /* length of received sequences */
#define AGE_NOT_SPECIFIED  0xFFFF
#define DEFAULT_ICQ_PORT 4000   /* ICQ UDP port number */
#define DEFAULT_ICQ_HOST "205.188.153.104"   /* XXX use dnsrv to resolve icq.mirabilis.com */
//#define PACKET_DEBUG

typedef struct session_st *session, _session;

/* ICQ Transport Instance */
typedef struct
{
    instance i;
    xdbcache xc;
    xmlnode vcard;
    HASHTABLE sessions;
    session session_list;
    int sessions_count;
    char *reg;
    char *server;
    char *search_msg, *search_timeout;
    int disable_tcp;
    int port, port_high, port_low;
    int start;
    int shutdown;
    pth_cond_t *shutdown_cond;
} *iti, _iti;

typedef enum
{
    genderMale,                   /* snakes 'n' snails 'n' puppy dog tails */
    genderFemale,                 /* sugar 'n' spice 'n' everything nice */
    genderNotSpecified            /* Chemical X :-) */
} icqgender;

typedef struct
{
    char *nick;           /* nickname */
    char *first;          /* first name */
    char *last;           /* last name */
    char *email;          /* primary email address */
    char *email2;         /* secondary email address */
    char *old_email;      /* old email address */
    char *city;           /* city */
    char *state;          /* state/province */
    char *phone;          /* phone number */
    char *fax;            /* fax number */
    char *street;         /* street address  */
    char *cellphone;      /* cellphone number  */
    unsigned long zipcode;      /* zip code (how shortsighted!) */
    unsigned short country_code; /* country code */
    int tz_offset;              /* timezone offset */
    int auth_required;          /* authorization required to add to contacts? */
    int web_aware;              /* publish my online status to Web as well? */
    int hide_ip;                /* hide my IP address from viewers? */
    /* -- from META_USER_INFO_2 */
    unsigned short age;         /* age */
    icqgender gender;           /* gender */
    char *homepage;       /* homepage URL  */
    unsigned short b_year;      /* birth year */
    unsigned short b_month;     /* birth month */
    unsigned short b_day;       /* birth day */
    unsigned short languages[3]; /* spoken language codes */
    /* -- from META_USER_INFO_WORK */
    char *w_city;         /* work city  */
    char *w_state;        /* work state/province */
    char *w_phone;        /* work phone number */
    char *w_fax;          /* work fax number */
    char *w_street;       /* work street address  */
    unsigned long w_zipcode;    /* zip code (how shortsighted!) */
    unsigned short w_country_code; /* country code */
    char *w_company;      /* company name */
    char *w_dept;         /* department name */
    char *w_title;        /* position title */
    unsigned short w_job;       /* job classification code */
    char *w_homepage;     /* work homepage URL  */
    /* -- from META_USER_INFO_ABOUT */
    char *profile;        /* user profile string */

    UIN_t uin; /* for search */
} *info_meta, _info_meta;

typedef struct
{
    pool p;
    info_meta meta;
    void *cb, *arg;
} *pendmeta, _pendmeta;

typedef enum
{
    STATUS_NOT_IN_LIST = -3,
    STATUS_OFFLINE = -1,
    STATUS_ONLINE = 0,
    STATUS_AWAY,
    STATUS_DND,
    STATUS_NA,
    STATUS_OCCUPIED,
    STATUS_FREE_CHAT,
    STATUS_INVISIBLE
} icqstatus;

typedef struct contact_st
{
    pool p;
    session s;
    UIN_t uin;
    icqstatus status;

    mio m;
    pth_msgport_t queue;
    unsigned connected;

    unsigned force_server;
    struct sockaddr_in addr;    /* peer address (LAN External) and port to connect to */
    struct in_addr addr_real;   /* contact's "real" IP address (LAN Internal) */
    unsigned long tcp_seq;      /* pending TCP sequence */
    unsigned short tcp_type;    /* pending TCP message type (used with CANCEL) */
    int port;

    int asking;
    int subscribing;

    struct contact_st *next;
} *contact, _contact;

typedef struct pqueue_st
{
    pool p;

    unsigned short seq1;
    time_t exp_time;
    int attempts;

    void *packet;
    unsigned short len;

    struct pqueue_st *next;
} *pqueue, _pqueue;

typedef enum {stype_normal, stype_register, stype_register_new} stype;

struct session_st
{
    pool p;
    jid id;
    jid from;
    mtq q;
    iti ti;
    stype type;
    UIN_t uin;
    ppdb p_db;
    pth_msgport_t queue;
    icqstatus status;
    char *status_text;
    int connected;
    int exit_flag;

    char *passwd;
    char *nick;
    char *first;
    char *last;
    char *email;

    mio server;
    mio peer_accept;

    contact contacts;
    int contacts_count;

    pendmeta pend_info;
    pendmeta pend_search;

    pqueue pq_head, pq_tail;
    time_t timer;
    int rsend;

    unsigned long sid;          /* session ID value (constant for connection) */
    unsigned short seq1;        /* sequence 1 value (starts at random) */
    unsigned short seq2;        /* sequence 2 value (starts at 1) */
    unsigned long tcp_gen;      /* TCP sequence generator for output */

    struct sockaddr_in addr;    /* my IP address (LAN Internal), and port of my listening socket */
    struct in_addr addr_s;      /* what the server thinks our IP address is (LAN External) */

    unsigned short rseq[RSEQ_MAX]; /* received sequence numbers */
    int rseq_cur;

    struct session_st *next;
    struct session_st *prev;
};

typedef struct
{
    pth_message_t head;
    jpacket jp;
} *jpq, _jpq;

typedef struct
{
    pth_message_t head;
    pool p;
    xmlnode x;
    UIN_t to;
    unsigned short type;
    char **parts;
} *imessage, _imessage;

typedef void (*info_meta_cb)(session s, info_meta info, void *arg);

void it_deliver(iti ti, xmlnode x);
session it_session_create(iti ti, jpacket jp);
void it_session_start(void *arg);
void it_session_jpacket(session s, jpacket jp);
void it_session_register(void *arg);
void it_session_register_fail(session s, terror e);
void it_session_end(session s);

void it_unknown(iti ti, jpacket jp);
void it_message(session s, jpacket jp);
void it_iq(session s, jpacket jp);
void it_iq_vcard_server(iti ti, jpacket jp);
void it_iq_browse_server(iti ti, jpacket jp);
void it_iq_time(iti ti, jpacket jp);
void it_iq_version(iti ti, jpacket jp);
void it_iq_last(iti ti, jpacket jp);
void it_presence(session s, jpacket jp);
void it_s10n(session s, jpacket jp);

void it_contact_load_roster(session s);
contact it_contact_get(session s, UIN_t uin);
contact it_contact_add(session s, UIN_t uin);
void it_contact_remove(contact c);
void it_contact_update_addresses(contact c, unsigned long ip,unsigned short port, unsigned long real_ip, int force_server);
void it_contact_set_status(contact c, icqstatus status);
void it_contact_send_presence(contact c);
void it_contact_subscribe(void *arg);
void it_contact_unsubscribe(void *arg);

int it_peer_listen(session s);
void it_peer_connect(contact c);
void it_peer_close(contact c);
void it_peer_process(contact c, icqbyte *buffer);
void it_peer_send_message(contact c, imessage m);

pendmeta it_meta_new(session s, info_meta_cb cb, void *arg);
void it_meta_packet(session s, PCICQDECODE packet);

int it_server_connect(session s);
void it_server_write(session s, void *buf, unsigned len, unsigned short cmd,  unsigned short seq1);
void it_build_and_write(session s, PICQBUFFER5C packet, unsigned len, unsigned short cmd);
void it_server_retransmit(session s);
void it_server_acked_packet(session s, unsigned short seq1);
void it_server_process_packet(session s, PCICQDECODE packet);

void it_server_login(session s);
void it_server_login_1(session s);
void it_server_new_UIN(session s);
void it_server_ack_messages(session s);
void it_server_text_code(session s, const char *code);
void it_server_keepalive(session s);
void it_server_send_contact_list(session s);
void it_server_send_status(session s);
void it_server_request_meta_info(session s, UIN_t uin);
void it_server_send_info(session s);
void it_server_search_uin(session s, UIN_t uin);
int it_server_search_email(session s, char *email);
int it_server_search_names(session s, char *nick, char *first, char *last);
int it_server_send_message(session s, imessage m);
int it_server_send_message_raw(session s, UIN_t uin, unsigned short type, unsigned len, icqbyte *data);

void it_icq_message(session s, UIN_t uin, char *ts, unsigned short type, icqbyte *data, unsigned short len);
imessage it_message_create(xmlnode x, UIN_t uin, unsigned short type, ...);
void it_message_send(contact c, imessage);
int it_message_size(imessage m);
icqbyte *it_message_format(imessage m, icqbyte *buffer);
int it_message_truncate(imessage m);

char *convert_utf8_to_windows(pool p, const char *utf8_str);
char *convert_windows_to_utf8(pool p, const char *windows_str);
int get_utf8_to_windows_len(const char *utf8_str);

void it_check_endian(void);
jid it_xdb_id(pool p, jid id, char *server);
UIN_t it_strtouin_safe(char *uin);
jid it_uin2jid(pool p, UIN_t uin, char *host);
icqstatus it_show2status(const char *show);
char *it_strrepl(pool p, const char *orig, const char *find, const char *replace);
void put_icqshort(void *buffer, unsigned offset, unsigned short value);
unsigned short get_icqshort(const void *buffer, unsigned offset);
unsigned long get_icqlong(const void *buffer, unsigned offset);
void put_icqlong(void *buffer, unsigned offset, unsigned long value);
unsigned long it_status_to_bits(icqstatus status);
icqstatus it_status_from_bits(unsigned long bits);
const char *it_country_from_code(unsigned short code);
inline icqshort to_icqshort(unsigned short arg);
inline icqlong to_icqlong(unsigned long arg);
inline unsigned long from_icqlong(icqlong arg);
inline unsigned short from_icqshort(icqshort arg);
unsigned short it_status_to_tcp(icqstatus status);
unsigned short it_format_tcp_msgcommand(unsigned short base_cmd, icqstatus status);
void it_debug_dump(const char *zone, const void *buffer, unsigned nbytes);
unsigned long it_checkvalue5(const icqbyte *packet,unsigned packet_len);
void it_scramble5(void *buffer, unsigned len, unsigned long checkvalue);
int it_test_v5_server(session s, void *buffer, unsigned len);
void it_delay(xmlnode x, char *ts);

#define it_jid2uin(x) it_strtouin_safe(x->user)
#define it_deliver(ti, x) deliver(dpacket_new(x),ti->i)
#define s_CRLF "\r\n"
#define s_LF "\n"
