/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "icqtransport.h"
#include <sys/utsname.h>

void it_iq_reg_get(void *arg);
void it_iq_reg_set(session s, jpacket jp);
void it_iq_search_get(session s, jpacket jp);
void it_iq_search_set(session s, jpacket jp);
void it_iq_vcard(session s, jpacket jp);
void it_iq_vcard_meta(session s, info_meta info, void *arg);
void it_iq_gateway_get(session s, jpacket jp);
void it_iq_gateway_set(session s, jpacket jp);
void it_iq_browse_user(session s, jpacket jp);

void it_iq(session s, jpacket jp)
{
    char *ns;

    if (s->connected == 0)
    {
        jpq jq;
        log_debug(ZONE,"Queued");
        jq = pmalloco(jp->p,sizeof(_jpq));
        jq->jp = jp;
        pth_msgport_put(s->queue,(pth_message_t *)jq);
        return;
    }

    ns = xmlnode_get_attrib(jp->iq,"xmlns");
    switch (jpacket_subtype(jp))
    {
    case JPACKET__GET:
        if (j_strcmp(ns,NS_REGISTER) == 0)
        {
            jp->aux1 = (void *) s;
            mtq_send(s->q,xmlnode_pool(jp->x),it_iq_reg_get,(void *) jp);
        }
        else if (j_strcmp(ns,NS_SEARCH) == 0)
            it_iq_search_get(s,jp);
        else if (j_strcmp(ns,NS_VERSION) == 0)
            it_iq_version(s->ti,jp);
        else if (j_strcmp(ns,NS_TIME) == 0)
            it_iq_time(s->ti,jp);
        else if (j_strcmp(ns,NS_VCARD) == 0)
            jp->to->user ? it_iq_vcard(s,jp) : it_iq_vcard_server(s->ti,jp);
        else if (j_strcmp(ns,NS_LAST) == 0)
            jp->to->user == NULL ? it_iq_last(s->ti,jp) : xmlnode_free(jp->x);
        else if (j_strcmp(ns,NS_BROWSE) == 0)
            jp->to->user ? it_iq_browse_user(s,jp) : it_iq_browse_server(s->ti,jp);
        else if (j_strcmp(ns,NS_GATEWAY) == 0)
            it_iq_gateway_get(s,jp);
        else
        {
            jutil_error(jp->x,TERROR_NOTIMPL);
            it_deliver(s->ti,jp->x);
        }
        break;

    case JPACKET__SET:
        if (j_strcmp(ns,NS_REGISTER) == 0)
            it_iq_reg_set(s,jp);
        else if (j_strcmp(ns,NS_SEARCH) == 0)
            it_iq_search_set(s,jp);
        else if (j_strcmp(ns,NS_GATEWAY) == 0)
            it_iq_gateway_set(s,jp);
        else
        {
            if (j_strcmp(ns,NS_VERSION) == 0 || j_strcmp(ns,NS_TIME) == 0)
                jutil_error(jp->x,TERROR_NOTALLOWED);
            else
                jutil_error(jp->x,TERROR_NOTIMPL);  
            it_deliver(s->ti,jp->x);
        }
        break;

    default:
        xmlnode_free(jp->x);
        break;
    }
}

void it_iq_reg_get(void *arg)
{
    jpacket jp = (jpacket) arg;
    session s = (session) jp->aux1;
    iti ti = s->ti;
    xmlnode q, reg;

    reg = xdb_get(ti->xc,it_xdb_id(xmlnode_pool(jp->x),s->id,s->from->server),NS_REGISTER);
    if (reg == NULL)
    {
        jutil_error(jp->x,TERROR_UNAVAIL);
        it_deliver(ti,jp->x);
        return;
    }

    jutil_iqresult(jp->x);
    q = xmlnode_insert_tag(jp->x,"query");
    xmlnode_put_attrib(q,"xmlns",NS_REGISTER);
    xmlnode_insert_node(q,xmlnode_get_firstchild(reg));
    xmlnode_free(reg);

    if (xmlnode_get_tag(q,"nick") == NULL)
        xmlnode_insert_tag(q,"nick");

    if (xmlnode_get_tag(q,"first") == NULL)
        xmlnode_insert_tag(q,"first");

    if (xmlnode_get_tag(q,"last") == NULL)
        xmlnode_insert_tag(q,"last");

    if (xmlnode_get_tag(q,"email") == NULL)
        xmlnode_insert_tag(q,"email");

    xmlnode_hide(xmlnode_get_tag(q,"password"));
    xmlnode_insert_tag(q,"password");

    xmlnode_insert_cdata(xmlnode_insert_tag(q,"key"),jutil_regkey(NULL,jid_full(jp->from)),-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(q,"instructions"),ti->reg,-1);

    it_deliver(ti,jp->x);
}

void _it_iq_reg_null(void *arg)
{
    session s = (session) arg;
    iti ti = s->ti;
    jid id;

    id = it_xdb_id(s->p,s->id,s->from->server);
    xdb_set(ti->xc,id,NS_REGISTER,NULL);
    xdb_set(ti->xc,id,NS_ROSTER,NULL);
}
    
void it_iq_reg_set(session s, jpacket jp)
{
    xmlnode query = jp->iq;
    xmlnode x;
    char *key;

    /* Verify that the registration key is correct. */
    key = xmlnode_get_tag_data(query,"key");
    if (key == NULL || !jutil_regkey(key,jid_full(jp->from)))
    {
        jutil_error(jp->x,TERROR_BAD);
        it_deliver(s->ti,jp->x);
        return;
    }

    if (xmlnode_get_tag(query,"remove"))
    {
        /* XXX Send an unsubscribe from all contacts in the contact list. */

        /* send an unsubscribe for the transport JID to the user */
        x = jutil_presnew(JPACKET__UNSUBSCRIBE,jid_full(s->id),NULL);
        xmlnode_put_attrib(x,"from",jid_full(s->from));
        it_deliver(s->ti,x);

        jutil_iqresult(jp->x);
        it_deliver(s->ti,jp->x);

        mtq_send(s->q,s->p,_it_iq_reg_null,(void *) s);
        it_session_end(s);
    }
    else
    {
        /* XXX */
        jutil_error(jp->x,TERROR_NOTIMPL);
        it_deliver(s->ti,jp->x);
    }
}

void it_iq_search_get(session s, jpacket jp)
{
    iti ti = s->ti;
    xmlnode q;

    if (jp->to->user != NULL)
    {
        jutil_error(jp->x,TERROR_NOTALLOWED);  
        it_deliver(s->ti,jp->x);
        return;
    }

    jutil_iqresult(jp->x);
    q = xmlnode_insert_tag(jp->x,"query");
    xmlnode_put_attrib(q,"xmlns",NS_SEARCH);

    /* Fill in the values on which we can search. */
    xmlnode_insert_tag(q,"username");
    xmlnode_insert_tag(q,"email");
    xmlnode_insert_tag(q,"nick");
    xmlnode_insert_tag(q,"first");
    xmlnode_insert_tag(q,"last");

    xmlnode_insert_cdata(xmlnode_insert_tag(q,"instructions"),ti->search_msg,-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(q,"key"),jutil_regkey(NULL,jid_full(jp->from)),-1);

    it_deliver(ti,jp->x);
}

void it_iq_search_result(session s, info_meta info, void *arg)
{
    jpacket jp = (jpacket) arg;

    if (info)
    {
        xmlnode item;
        pool p = jp->p;

        item = xmlnode_insert_tag(jp->iq,"item");

        xmlnode_put_attrib(item,"jid",jid_full(it_uin2jid(p,info->uin,s->from->server)));

        xmlnode_insert_cdata(xmlnode_insert_tag(item,"email"),
                             convert_windows_to_utf8(p,info->email),-1);

        xmlnode_insert_cdata(xmlnode_insert_tag(item,"nick"),
                             convert_windows_to_utf8(p,info->nick),-1);

        xmlnode_insert_cdata(xmlnode_insert_tag(item,"first"),
                             convert_windows_to_utf8(p,info->first),-1);

        xmlnode_insert_cdata(xmlnode_insert_tag(item,"last"),
                             convert_windows_to_utf8(p,info->last),-1);
    }
    else
        it_deliver(s->ti,jp->x);
}

void it_iq_search_set(session s, jpacket jp)
{
    xmlnode query = jp->iq;
    pool p;
    UIN_t uin;
    char *first, *last, *nick, *email, *key;

    if (s->pend_search != NULL || jp->to->user != NULL)
    {
        jutil_error(jp->x,TERROR_NOTALLOWED);
        it_deliver(s->ti,jp->x);
        return;
    }

    /* Verify that the registration key is correct. */
    key = xmlnode_get_tag_data(query,"key");
    if (key == NULL || jutil_regkey(key,jid_full(jp->from)) == NULL)
    {
        jutil_error(jp->x,TERROR_BAD);
        it_deliver(s->ti,jp->x);
        return;
    }

    p = jp->p;
    uin = it_strtouin_safe(xmlnode_get_tag_data(query,"username"));
    nick = convert_utf8_to_windows(p,xmlnode_get_tag_data(query,"nick"));
    first = convert_utf8_to_windows(p,xmlnode_get_tag_data(query,"first"));
    last = convert_utf8_to_windows(p,xmlnode_get_tag_data(query,"last"));
    email = convert_utf8_to_windows(p,xmlnode_get_tag_data(query,"email"));

    /* Make sure at least ONE of the possible search parameters is specified */
    if (uin == 0 && nick == NULL && first == NULL && last == NULL && email == NULL)
    {
        /* no parameters specified */
        jutil_error(jp->x,TERROR_BAD);
        return;
    }

    jutil_iqresult(jp->x);
    xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"),"xmlns",NS_SEARCH);
    jpacket_reset(jp);

    s->pend_search = it_meta_new(s,it_iq_search_result,(void *) jp);

    /* Determine the search type. */
    if (uin != 0)
    {
        log_debug(ZONE,"search by uin");
        it_server_search_uin(s,uin);
    }
    else if (email)
    {
        log_debug(ZONE,"search by email");
        it_server_search_email(s,email);
    }
    else
        it_server_search_names(s,nick ? nick : "",first ? first : "",last ? last : "");
}

void it_iq_gateway_get(session s, jpacket jp)
{
    if (jp->to->user == NULL)
    {
        xmlnode q;

        jutil_iqresult(jp->x);

        q = xmlnode_insert_tag(jp->x,"query");
        xmlnode_put_attrib(q,"xmlns",NS_GATEWAY);
        xmlnode_insert_cdata(xmlnode_insert_tag(q,"desc"),"Enter the user's UIN",-1);
        xmlnode_insert_tag(q,"prompt");
    }
    else
        jutil_error(jp->x,TERROR_NOTALLOWED);

    it_deliver(s->ti,jp->x);
}

void it_iq_gateway_set(session s, jpacket jp)
{
    char *user, *id;

    user = xmlnode_get_tag_data(jp->iq,"prompt");
    id = user ? spools(jp->p,user,"@",jp->to->server,jp->p) : NULL;
    if (id)
    {
        xmlnode q;

        jutil_iqresult(jp->x);
        
        q = xmlnode_insert_tag(jp->x,"query");
        xmlnode_put_attrib(q,"xmlns",NS_GATEWAY);
        xmlnode_insert_cdata(xmlnode_insert_tag(q,"prompt"),id,-1);
    }
    else
        jutil_error(jp->x,TERROR_BAD);

    it_deliver(s->ti,jp->x);
}

void it_iq_browse_server(iti ti, jpacket jp)
{
    xmlnode q;

    q = xmlnode_insert_tag(jutil_iqresult(jp->x),"service");
    xmlnode_put_attrib(q,"xmlns",NS_BROWSE);

    xmlnode_put_attrib(q,"type","icq");
    xmlnode_put_attrib(q,"jid",jp->to->server);
    xmlnode_put_attrib(q,"name",xmlnode_get_tag_data(ti->vcard,"FN"));

    xmlnode_insert_cdata(xmlnode_insert_tag(q,"ns"),NS_REGISTER,-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(q,"ns"),NS_SEARCH,-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(q,"ns"),NS_GATEWAY,-1);

    it_deliver(ti,jp->x);
}

void it_iq_browse_user(session s, jpacket jp)
{
    xmlnode browse;

    if (s->type == stype_register)
    {
        jpq jq;

        jq = pmalloco(jp->p,sizeof(_jpq));
        jq->jp = jp;
        pth_msgport_put(s->queue,(pth_message_t *)jq);
        return;
    }

    if (it_jid2uin(jp->from) == 0)
    {
        jutil_error(jp->x,TERROR_BAD);
        it_deliver(s->ti,jp->x);
        return;
    }

    jutil_iqresult(jp->x);
    browse = xmlnode_insert_tag(jp->x,"user");
    xmlnode_put_attrib(browse,"jid",jid_full(jid_user(jp->to)));
    xmlnode_put_attrib(browse,"type","user");

    it_deliver(s->ti,jp->x);
}

void it_iq_version(iti ti, jpacket jp)
{
    xmlnode os, x, q;
    struct utsname un;

    x = jutil_iqresult(jp->x);

    q = xmlnode_insert_tag(x,"query");
    xmlnode_put_attrib(q,"xmlns",NS_VERSION);

    xmlnode_insert_cdata(xmlnode_insert_tag(q,"name"),"ICQ Transport",-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(q,"version"),"0.9",-1);

    uname(&un);
    os = xmlnode_insert_tag(q,"os");
    xmlnode_insert_cdata(os,un.sysname,-1);
    xmlnode_insert_cdata(os," ",1);
    xmlnode_insert_cdata(os,un.release,-1);

    it_deliver(ti,x);
}

void it_iq_time(iti ti, jpacket jp)
{
    xmlnode x, q;

    x = jutil_iqresult(jp->x);

    q = xmlnode_insert_tag(x,"query");
    xmlnode_put_attrib(q,"xmlns",NS_TIME);

    xmlnode_insert_cdata(xmlnode_insert_tag(q,"utc"),jutil_timestamp(),-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(q,"tz"),tzname[0],-1);

    it_deliver(ti,x);
}

void it_iq_last(iti ti, jpacket jp)
{
    xmlnode x, q;
    char str[10];

    x = jutil_iqresult(jp->x);

    snprintf(str,10,"%d",time(NULL) - ti->start);
    q = xmlnode_insert_tag(x,"query");
    xmlnode_put_attrib(q,"xmlns",NS_LAST);
    xmlnode_put_attrib(q,"seconds",str);

    it_deliver(ti,x);
}

void it_iq_vcard_server(iti ti, jpacket jp)
{
    /* send back transport vCard */
    xmlnode_insert_node(jutil_iqresult(jp->x),ti->vcard);
    it_deliver(ti,jp->x);
}

void it_iq_vcard(session s, jpacket jp)
{
    UIN_t uin;
    xmlnode data;

    if (s->pend_info != NULL)
    {
        jutil_error(jp->x,TERROR_NOTALLOWED);
        it_deliver(s->ti,jp->x);
        return;
    }

    uin = it_jid2uin(jp->to);
    if (uin == 0)
    {
        jutil_error(jp->x,TERROR_BAD);
        it_deliver(s->ti,jp->x);
        return;
    }

    jutil_iqresult(jp->x);
    jp->iq = data = xmlnode_insert_tag(jp->x,"vCard");
    xmlnode_put_attrib(data,"xmlns",NS_VCARD);
    xmlnode_put_attrib(data,"version","3.0");
    xmlnode_put_attrib(data,"prodid","-//HandGen//NONSGML vGen v1.0//EN");

    /* Request meta-info for the user. */
    s->pend_info = it_meta_new(s,it_iq_vcard_meta,(void *) jp);
    it_server_request_meta_info(s,uin);
}

#define VCARD_ADD(q, parent, name, value) \
    xmlnode_insert_cdata(xmlnode_insert_tag(xmlnode_insert_tag(q,parent),name), \
    convert_windows_to_utf8(xmlnode_pool(q),value),-1)

void it_iq_vcard_meta(session s, info_meta info, void *arg)
{
    jpacket jp = (jpacket) arg;
    xmlnode q, t;
    pool p;
    const char *c;
    char buffer[16];

    if (info == NULL)
    {
        it_deliver(s->ti,jp->x);
        return;
    }

    q = jp->iq;
    p = jp->p;

    t = xmlnode_insert_tag(q,"FN");
    if (info->first)
    {
        /* we have a first name - if we have a last name, we have to concatenate it */
        if (info->last)
            xmlnode_insert_cdata(t,convert_windows_to_utf8(p,spools(p,info->first," ",info->last,p)),-1);
        else
            xmlnode_insert_cdata(t,convert_windows_to_utf8(p,info->first),-1);
    }
    else if (info->last)
        xmlnode_insert_cdata(t,convert_windows_to_utf8(p,info->last),-1);

    t = xmlnode_insert_tag(q,"N");
    if (info->first)
        xmlnode_insert_cdata(xmlnode_insert_tag(t,"GIVEN"),
                             convert_windows_to_utf8(p,info->first),-1);
    if (info->last)
        xmlnode_insert_cdata(xmlnode_insert_tag(t,"FAMILY"),
                             convert_windows_to_utf8(p,info->last),-1);

    if (info->nick)
        xmlnode_insert_cdata(xmlnode_insert_tag(q,"NICKNAME"),
                             convert_windows_to_utf8(p,info->nick),-1);

    if (info->email)
    {
        t = VCARD_ADD(q,"EMAIL","USERID",info->email);
        xmlnode_insert_tag(t,"INTERNET");
        xmlnode_insert_tag(t,"PREF");
    }

    if (info->email2)
    {
        t = VCARD_ADD(q,"EMAIL","USERID",info->email2);
        xmlnode_insert_tag(t,"INTERNET");    
    }

    if (info->homepage)
        xmlnode_insert_cdata(xmlnode_insert_tag(q,"URL"),convert_windows_to_utf8(p,info->homepage),-1);

    if (info->phone)
    {
        t = VCARD_ADD(q,"TEL","NUMBER",info->phone);
        xmlnode_insert_tag(t,"VOICE");
        xmlnode_insert_tag(t,"HOME");
    }

    if (info->fax)
    {
        t = VCARD_ADD(q,"TEL","NUMBER",info->fax);
        xmlnode_insert_tag(t,"FAX");
        xmlnode_insert_tag(t,"HOME");
    }

    if (info->cellphone)
    {
        t = VCARD_ADD(q,"TEL","NUMBER",info->cellphone);
        xmlnode_insert_tag(t,"VOICE");
        xmlnode_insert_tag(t,"CELL");   
    }

    t = xmlnode_insert_tag(q,"ADR");
    xmlnode_insert_tag(t,"HOME");
    xmlnode_insert_tag(t,"EXTADD");

    if (info->street)
        xmlnode_insert_cdata(xmlnode_insert_tag(t,"STREET"),convert_windows_to_utf8(p,info->street),-1);

    if (info->city)
        xmlnode_insert_cdata(xmlnode_insert_tag(t,"LOCALITY"),convert_windows_to_utf8(p,info->city),-1);

    if (info->state)
        xmlnode_insert_cdata(xmlnode_insert_tag(t,"REGION"),convert_windows_to_utf8(p,info->state),-1);

    if (info->zipcode != 0)
    {
        snprintf(buffer,16,"%05lu",info->zipcode);
        xmlnode_insert_cdata(xmlnode_insert_tag(t,"PCODE"),buffer,-1);   
    }

    c = it_country_from_code(info->country_code);
    xmlnode_insert_cdata(xmlnode_insert_tag(t,"COUNTRY"),c,-1);

    if (info->w_company)
    {
        t = VCARD_ADD(q,"ORG","ORGNAME",info->w_company);
        if (info->w_dept)
            xmlnode_insert_cdata(xmlnode_insert_tag(t,"ORGUNIT"),
                                 convert_windows_to_utf8(p,info->w_dept),-1);
    }

    if (info->w_phone)
    {
        t = VCARD_ADD(q,"TEL","NUMBER",info->w_phone);
        xmlnode_insert_tag(t,"VOICE");
        xmlnode_insert_tag(t,"WORK");
    }

    if (info->w_fax)
    {
        t = VCARD_ADD(q,"TEL","NUMBER",info->w_fax);
        xmlnode_insert_tag(t,"FAX");
        xmlnode_insert_tag(t,"WORK");
    }

    t = xmlnode_insert_tag(q,"ADR");
    xmlnode_insert_tag(t,"WORK");
    xmlnode_insert_tag(t,"EXTADD");

    if (info->w_street)
        xmlnode_insert_cdata(xmlnode_insert_tag(t,"STREET"),convert_windows_to_utf8(p,info->w_street),-1);

    if (info->w_city)
        xmlnode_insert_cdata(xmlnode_insert_tag(t,"LOCALITY"),convert_windows_to_utf8(p,info->w_city),-1);

    if (info->w_state)
        xmlnode_insert_cdata(xmlnode_insert_tag(t,"REGION"),convert_windows_to_utf8(p,info->w_state),-1);

    if (info->w_zipcode != 0)
    {
        snprintf(buffer,16,"%05lu",info->w_zipcode);
        xmlnode_insert_cdata(xmlnode_insert_tag(t,"PCODE"),buffer,-1);
    }

    c = it_country_from_code(info->w_country_code);
    xmlnode_insert_cdata(xmlnode_insert_tag(t,"COUNTRY"),c,-1);

    if (info->profile)
        xmlnode_insert_cdata(xmlnode_insert_tag(q,"DESC"),info->profile,-1);

    it_deliver(s->ti,jp->x);

    /* N.B.: we do NOT use old_email */
    /* N.B.: we do NOT use tz_offset */
    /* N.B.: we do NOT use auth_required */
    /* N.B.: we do NOT use web_aware */
    /* N.B.: we do NOT use hide_ip */
    /* N.B.: we do NOT use age */
    /* N.B.: we do NOT use gender */
    /* N.B.: we do NOT use languages */
    /* N.B.: we do NOT use w_title */
    /* N.B.: we do NOT use w_job */
    /* N.B.: we do NOT use w_homepage */
    /* FUTURE: format in birthday (b_year, b_month, b_day)? */
}
