/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "icqtransport.h"

void it_message_roster(session s, contact c, jpacket jp, int count);

void it_message(session s, jpacket jp)
{
    UIN_t uin;
	contact c;
    xmlnode  sub, item;
    char *body;
    int count = 0;

    if (s->connected == 0)
    {
        jpq jq;
        log_debug(ZONE,"Queued");
        jq = pmalloco(jp->p,sizeof(_jpq));
        jq->jp = jp;
        pth_msgport_put(s->queue,(pth_message_t *)jq);
        return;
    }

    uin = it_jid2uin(jp->to);
    if (uin == 0 || s->uin == uin)
    {
        jutil_error(jp->x,TERROR_BAD);
        it_deliver(s->ti,jp->x);
        return;      
    }

    c = it_contact_get(s,uin);
	if (c == NULL)
        c = it_contact_add(s,uin); /* why not? */

    /* count x:roster items, if any */
    for (item = xmlnode_get_firstchild(xmlnode_get_tag(jp->x,"x?xmlns=jabber:x:roster"));
         item; item = xmlnode_get_nextsibling(item)) {

        if (j_strcmp(xmlnode_get_name(item),"item") == 0)
            count++;
    }

    if (count)
    {
        it_message_roster(s,c,jp,count);
        return;
    }

    if ((sub = xmlnode_get_tag(jp->x,"x?xmlns=jabber:x:oob")) != NULL)
    {
        char *url;

        url = xmlnode_get_tag_data(sub,"url");
        if (url)
        {
            imessage m;

            m = it_message_create(jp->x,c->uin,MSGTYPE_URL,xmlnode_get_tag_data(sub,"desc"),url);
            if (m == NULL)
            {
                jutil_error(jp->x,(terror){400,"Failed to convert UTF-8"});
                it_deliver(s->ti,jp->x);
            }
            else
                it_message_send(c,m);
            return;
        }
    }

    body = xmlnode_get_tag_data(jp->x,"body");
    if (body == NULL)
    {
        jutil_error(jp->x,TERROR_BAD);
        it_deliver(s->ti,jp->x);
    }
    else
    {
        imessage m;

        m = it_message_create(jp->x,c->uin,MSGTYPE_TEXT,body);
        if (m == NULL)
        {
            jutil_error(jp->x,(terror){400,"Failed to convert from UTF-8"});
            it_deliver(s->ti,jp->x);
        }
        else
            it_message_send(c,m);
    }
}

void it_message_roster(session s, contact c, jpacket jp, int count)
{
    imessage m;
    xmlnode item;
    jid id;
    char **parts;
    int i = 0;

    log_debug(ZONE,"Session[%s], roster message",jid_full(s->id));

    parts = (char **) pmalloco(jp->p,sizeof(char *) * (count * 2));

    for (item = xmlnode_get_firstchild(xmlnode_get_tag(jp->x,"x?xmlns=jabber:x:roster"));
         item != NULL; item = xmlnode_get_nextsibling(item)) {

        if (j_strcmp(xmlnode_get_name(item),"item") != 0) continue;

        id = jid_new(jp->p,xmlnode_get_attrib(item,"jid"));
        parts[i] = id->user;

        ++i;

        parts[i] = xmlnode_get_attrib(item,"name");
        if (parts[i] == NULL)
            parts[i] = id->user;
        i++;
    }

    m = it_message_create(jp->x,c->uin,MSGTYPE_CONTACTS);
    m->parts = parts;
    it_message_send(c,m);
}
