/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "icqtransport.h"

void it_meta_info(session s, PCICQMETADECODE info);
void it_meta_search(session s, PCICQMETADECODE info, int last);

pendmeta it_meta_new(session s, info_meta_cb cb, void *arg)
{
    pool p;
    pendmeta meta;

    p = pool_new();
    meta = pmalloco(p,sizeof(_pendmeta));
    meta->p = p;
    meta->cb = cb;
    meta->arg = arg;
    meta->meta = pmalloco(p,sizeof(_info_meta));

    return meta;
}

void it_meta_packet(session s, PCICQDECODE packet)
{
    ICQMETADECODE mp;

    mp.base_info = packet;
    mp.meta_command = get_icqshort(packet->data,0);
    mp.data = packet->data + 3;
    mp.datalen = packet->datalen - 3;
    mp.success = (packet->data[2]==(icqbyte)0x0A);

    switch (mp.meta_command)
    {
    case META_USER_INFO:
    case META_USER_INFO_2:
    case META_USER_PICTURE:
    case META_USER_WORK:
    case META_USER_ABOUT:
    case META_USER_X1:
    case META_USER_INFO_EOF:
        it_meta_info(s,&mp);
        break;

    case META_USER_FOUND:
    case META_LAST_USER_FOUND:
        it_meta_search(s,&mp,(mp.meta_command == META_LAST_USER_FOUND));
        break;

    default:
        log_debug(ZONE,"(unknown meta-command data)");
        break;
    }
}

#define META_NEXT_STRING(str) \
    len = get_icqshort(block,offset); \
    offset += 2; \
    str = len != 1 ? (char *) (block + offset) : NULL;\
    offset += len; \

void it_meta_info(session s, PCICQMETADECODE info)
{
    pendmeta pend = s->pend_info;

    if (pend)
    {
        icqbyte *block;
        info_meta meta = pend->meta;
        unsigned offset = 0, len = 0;

        if (info->success == 0 || info->meta_command == META_USER_INFO_EOF)
        {
            log_debug(ZONE,"Session[%s], meta finished",jid_full(s->id));
            ((info_meta_cb) (pend->cb)) (s,meta,pend->arg);
            s->pend_info = NULL;
            pool_free(pend->p);
            return;
        }

        if (info->meta_command == META_USER_X1 || info->meta_command == META_USER_PICTURE)
        {
            /* we don't know this format */
            log_debug(ZONE,"META_USER (unknown format)");
            return;
        }


        block = (icqbyte *) pmalloco(pend->p,info->datalen);
        memcpy(block,info->data,info->datalen);

        switch (info->meta_command)
        {
        case META_USER_INFO:
            log_debug(ZONE,"META_USER_INFO, filling in");

            META_NEXT_STRING(meta->nick);
            META_NEXT_STRING(meta->first);
            META_NEXT_STRING(meta->last);
            META_NEXT_STRING(meta->email);
            META_NEXT_STRING(meta->email2);
            META_NEXT_STRING(meta->old_email);
            META_NEXT_STRING(meta->city);
            META_NEXT_STRING(meta->state);
            META_NEXT_STRING(meta->phone);
            META_NEXT_STRING(meta->fax);
            META_NEXT_STRING(meta->street);
            META_NEXT_STRING(meta->cellphone);

            meta->zipcode = get_icqlong(block,offset);
            meta->country_code = get_icqshort(block,offset + 4);
            meta->tz_offset = (int)(block[offset + 6]);
            meta->auth_required = (block[offset + 7] ? 0 : 1);
            meta->web_aware = ((block[offset + 8] == (icqbyte)1) ? 1 : 0);
            meta->hide_ip = (block[offset + 9] ? 1 : 0);
            break;

        case META_USER_INFO_2:
            log_debug(ZONE,"META_USER_INFO_2, filling in");
            meta->age = get_icqshort(block,0);
            if (block[2] == (icqbyte) 1)
                meta->gender = genderFemale;
            else if (block[2] == (icqbyte) 2)
                meta->gender = genderMale;
            else
                meta->gender = genderNotSpecified;

            offset = 3;
            META_NEXT_STRING(meta->homepage);

            meta->b_year = (unsigned short)(block[offset++]) + 1900;
            meta->b_month = (unsigned short)(block[offset++]);
            meta->b_day = (unsigned short)(block[offset++]);
            meta->languages[0] = (unsigned short)(block[offset++]);
            meta->languages[1] = (unsigned short)(block[offset++]);
            meta->languages[2] = (unsigned short)(block[offset]);
            break;

        case META_USER_WORK:
            log_debug(ZONE,"META_USER_WORK, filling in");

            META_NEXT_STRING(meta->w_city);
            META_NEXT_STRING(meta->w_state);
            META_NEXT_STRING(meta->w_phone);
            META_NEXT_STRING(meta->w_fax);
            META_NEXT_STRING(meta->w_street);

            meta->w_zipcode = get_icqlong(block,offset);
            meta->w_country_code = get_icqshort(block,offset + 4);
            offset += 6;

            META_NEXT_STRING(meta->w_company);
            META_NEXT_STRING(meta->w_dept);
            META_NEXT_STRING(meta->w_title);

            meta->w_job = get_icqshort(block,offset);
            meta->w_homepage = (char *)(block + offset + 4);
            break;

        case META_USER_ABOUT:
            log_debug(ZONE,"META_USER_ABOUT, filling in");
            meta->profile = (char *)(block + 2);
            break;

        default:
            log_warn(ZONE,"meta code %04X unknown",info->meta_command);
            break;       
        }
    }
}

void it_meta_search(session s, PCICQMETADECODE info, int last)
{
    pendmeta pend = s->pend_search;

    if (pend == NULL)
        return;

    log_debug(ZONE,"inforeq_receive_meta_search, %d",last);

    if (last == 0 && info->success)
    {
        info_meta meta = pend->meta;
        icqbyte *block = info->data;
        unsigned offset, len;

        /* Break up the received user information. */
        meta->uin = get_icqlong(block,0);
        offset = 4;

        META_NEXT_STRING(meta->nick);
        META_NEXT_STRING(meta->first);
        META_NEXT_STRING(meta->last);
        META_NEXT_STRING(meta->email);

        meta->auth_required = ((icqbyte) block[offset] == (icqbyte) 1) ? 0 : 1;

        log_debug(ZONE,"SEARCH RESULT: UIN %lu, nick %s, name %s %s, email %s",meta->uin,
                  meta->nick,meta->first,meta->last,meta->email);

        ((info_meta_cb) (pend->cb)) (s,meta,pend->arg);
    }
    else
    {
        ((info_meta_cb) (pend->cb)) (s,NULL,pend->arg);
        s->pend_search = NULL;
        pool_free(pend->p);
    }
}
