/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "icqtransport.h"

#define MIO_ICQ 6
typedef void (*mio_icq_cb)(mio m, int state, void *arg, icqbyte *buffer, icqshort size);

void it_peer_packet(mio m, int state, void *arg, icqbyte *buffer, icqshort size)
{
    contact c = (contact) arg;

    log_debug(ZONE,"peer packet %d",state);

    if (state == MIO_ICQ && size != 1 && c)
    {
        it_debug_dump(ZONE,buffer,size);
        it_peer_process(c,buffer);
    }
    else if ((state == MIO_CLOSED || state == MIO_ERROR) && c)
        it_peer_close(c);
}

void it_peer_accept(mio m, int state, void *arg, icqbyte *buffer, icqshort size)
{
    session s = (session) arg;

    if (s == NULL)
        return;

    switch(state)
    {
    case MIO_NEW:
        if (m == s->peer_accept)
            log_debug(ZONE,"accepted peer connection");
        break;
    case MIO_CLOSED:
    case MIO_ERROR:
        if (m == s->peer_accept)
        {
            log_debug(ZONE,"peer listening socket died");
        }
        break;
    case MIO_ICQ:
    {
        contact c;
        PICQTCP2INIT  packet = (PICQTCP2INIT) buffer;

        if (m == s->peer_accept || size == 1) return;

        if (packet->code != TCP_INIT_PACKET_CODE)
        {
            it_debug_dump(ZONE,buffer,size);
            log_debug(ZONE,"peer_accept: not an init packet");
            m->cb = NULL;
            mio_close(m);
            return;
        }

        c = it_contact_get(s,from_icqlong(packet->uin));
        if (c == NULL)
            c = it_contact_add(s,from_icqlong(packet->uin));

        log_debug(ZONE,"Session[%s], %lu initiating connection",jid_full(s->id),c->uin);

        c->connected = 1;
        c->force_server = 0;
        c->addr.sin_addr.s_addr = from_icqlong(packet->ip);
        c->addr.sin_port = htons(from_icqshort(packet->port));
        c->addr_real.s_addr = from_icqlong(packet->real_ip);
        c->m = m;
        m->cb = it_peer_packet;
        m->cb_arg = c;
    }
    }
}

void _it_peer_connect(mio m, int state, void *arg, icqbyte *buffer, icqshort size)
{
    contact c = (contact) arg;

    log_debug(ZONE,"Connect!");

    if (c == NULL)
        return;

    if (state == MIO_NEW)
    {
        ICQTCP2INIT packet;
        imessage msg;

        log_debug(ZONE,"Session[%s], sending peer init",jid_full(c->s->id));

        /* Fill in the data for the new init packet. */
        memset(&packet,0,sizeof(ICQTCP2INIT));
        packet.length = to_icqshort(sizeof(ICQTCP2INIT) - sizeof(icqshort));
        packet.code = TCP_INIT_PACKET_CODE;
        packet.version = to_icqshort(TCP_VERSION_NUMBER);
        packet.port = to_icqshort(ntohl(c->s->addr.sin_port));
        packet.uin = to_icqlong(c->s->uin);
        packet.ip = to_icqlong(c->s->addr_s.s_addr);
        packet.real_ip = to_icqlong(c->s->addr.sin_addr.s_addr);
        packet.four = (icqbyte)4;

        mio_write(c->m,NULL,(char *) &packet,sizeof(ICQTCP2INIT));

        m->cb = it_peer_packet;
        c->connected = 1;

        while ((msg = (imessage) pth_msgport_get(c->queue)) != NULL)
            it_peer_send_message(c,msg);
    }
    else
    {
        if (c->m == NULL)
        {
            log_debug(ZONE,"Failed to establish connection to peer");
        }
        else
            it_peer_close(c);
    }
}

ssize_t it_peer_read(mio m, void *unused, size_t bufsz)
{
    icqbyte *buf;
    icqshort size;
    int fd = m->fd;
    int len;

    /* Try to get the size of the received packet. */
    len = recv(fd,&size,sizeof(icqshort),MSG_PEEK);
    if (len <= 0)
        return len;
    else if (len < sizeof(icqshort))
        return 1;

    /* allocate a buffer for the packet and attempt to read it all in. */
    size = (int) from_icqshort(size) + sizeof(icqshort);
    buf = (icqbyte *) alloca(size);
    len = recv(fd,buf,size,MSG_PEEK);

    if (len <= 0)
        return len;
    else if (len < size)
        return 1;

    log_debug(ZONE,"packet length %d",size);

    len = recv(fd,buf,size,0);

    (*(mio_icq_cb)m->cb)(m,MIO_ICQ,m->cb_arg,buf,size);

    return 1;
}

int it_peer_listen(session s)
{
    mio m;
    iti ti = s->ti;
    int port = ti->port;

    /* Set up the range of ports to look in for a message. */
    if (ti->port > ti->port_high)
        port = ti->port = ti->port_low;

    port = ti->port++;

    log_debug(ZONE,"Port range = [%d,%d], start at %d",ti->port_low,ti->port_high,port);

    m = mio_listen(port,inet_ntoa(s->addr.sin_addr),it_peer_accept,(void *) s,NULL,
                   mio_handlers_new((mio_read_func) it_peer_read,NULL,NULL));

    if (m == NULL)
        return 1;

    s->addr.sin_port = htons((unsigned short) port);
    s->tcp_gen = 0xFFFFFFFEUL;

    return 0;
}

void it_peer_connect(contact c)
{
    mio_connect(inet_ntoa(c->addr.sin_addr),(int) c->port,_it_peer_connect,(void *) c,-1,NULL,
                mio_handlers_new(it_peer_read,NULL,NULL));
}

void it_peer_close(contact c)
{
    c->m->cb_arg = NULL;
    mio_close(c->m);
    c->m = NULL;
    c->connected = 0;
    c->tcp_seq = 0xFFFFFFFFUL;
    c->tcp_type = 0;
}
