/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "icqtransport.h"

void it_peer_ack_message(contact c, unsigned short type, unsigned long seq, char *response);

void it_peer_process(contact c, icqbyte *buffer)
{
    icqbyte *payload;      /* pointer to payload within packet */
    icqbyte *vtail;        /* virtual tail */
    PICQTCP2MSG1 pp1;      /* pointer to part 1 of packet */
    PICQTCP2MSG2 pp2;      /* pointer to part 2 of packet */
    UIN_t uin;             /* UIN of sender */
    PICQTCP2TLN ptl;       /* pointer to tail of packet */
    unsigned short command, type;
    unsigned payload_size;

    /* split the packet up into its various parts */
    pp1 = (PICQTCP2MSG1) buffer;

    payload = buffer + sizeof(ICQTCP2MSG1);
    payload_size = from_icqshort(pp1->textlen);

    vtail = payload + payload_size;
    pp2 = (PICQTCP2MSG2)vtail;
    vtail += sizeof(ICQTCP2MSG2);

    /* retrieve some information from the packet */
    uin = from_icqlong(pp1->uin);
    command = from_icqshort(pp1->command);
    type = from_icqshort(pp1->type) & ~MSGTYPE_FLAG_MASS;

    log_debug(ZONE,"incoming UIN %lu, command %04X, type %u",uin,command,type);

    if (uin != c->uin)
        log_debug(ZONE,"message came in on connection for %lu but with UIN %lu",c->uin,uin);

    switch (command)
    {
    case TCP_START:
        switch (type)
        {
        case MSGTYPE_CONTACT_ADD:
        case MSGTYPE_AUTH_REQUEST:
        case MSGTYPE_AUTHORIZE:
        case MSGTYPE_AUTH_DENY:
        case MSGTYPE_URL:
        case MSGTYPE_CONTACTS:
        case MSGTYPE_TEXT:
        case MSGTYPE_WEB_PAGER:
        case MSGTYPE_EMAIL_PAGER:
        case MSGTYPE_GREETING_CARD:
            ptl = (PICQTCP2TLN) vtail;
            it_peer_ack_message(c,type,from_icqlong(ptl->tcpseq),NULL);
            it_icq_message(c->s,uin,NULL,type,payload,(unsigned short) payload_size);
            break;

        case MSGTYPE_CHAT:
            /* FUTURE: this is where chat protocol would be implemented */
            log_debug(ZONE,"received chat request through TCP (ignored)");
            break;

        case MSGTYPE_FILE:
            /* FUTURE: this is where file direct protocol would be implemented */
            log_debug(ZONE,"received file request through TCP (ignored)");
            break;

        case MSGTYPE_READ_AWAY:
        case MSGTYPE_READ_OCC:
        case MSGTYPE_READ_NA:
        case MSGTYPE_READ_DND:
        case MSGTYPE_READ_FFC:
            ptl = (PICQTCP2TLN) vtail;
            it_peer_ack_message(c,type,from_icqlong(ptl->tcpseq),c->s->status_text);
            break;

        default:
            log_debug(ZONE,"unknown TCP_START message type %u",type);
            break;
        }
        break;

    case TCP_CANCEL:
        switch (type)
        {
        /* the cancel isn't really important at the moment */
        case MSGTYPE_CHAT:
            /* FUTURE: this is where chat protocol would be implemented */
            log_debug(ZONE,"received chat cancel through TCP (ignored)");
            break;

        case MSGTYPE_FILE:
            /* FUTURE: this is where file direct protocol would be implemented */
            log_debug(ZONE,"received file cancel through TCP (ignored)");
            break;

        default:
            break;
        }
        break;

    case TCP_ACK:
        c->tcp_seq = 0xFFFFFFFFUL;
        switch (type)
        {
        case MSGTYPE_CONTACT_ADD:
        case MSGTYPE_AUTH_REQUEST:
        case MSGTYPE_AUTHORIZE:
        case MSGTYPE_AUTH_DENY:
        case MSGTYPE_URL:
        case MSGTYPE_CONTACTS:
        case MSGTYPE_TEXT:
        case MSGTYPE_WEB_PAGER:
        case MSGTYPE_EMAIL_PAGER:
        case MSGTYPE_GREETING_CARD:
            /* FUTURE: this is where we read back autoresponse messages */
            break;

        case MSGTYPE_READ_AWAY:
        case MSGTYPE_READ_OCC:
        case MSGTYPE_READ_NA:
        case MSGTYPE_READ_DND:
        case MSGTYPE_READ_FFC:
            /* FUTURE: this is where we read back away messages */
            break;

        case MSGTYPE_CHAT:
            /* FUTURE: this is where chat protocol would be implemented */
            log_debug(ZONE,"received chat ack through TCP (ignored)");
            break;

        case MSGTYPE_FILE:
            /* FUTURE: this is where file direct protocol would be implemented */
            log_debug(ZONE,"received file ack through TCP (ignored)");
            break;

        default:
            log_debug(ZONE,"unknown TCP_ACK message type %u",type);
            break;
        }
        break;

    default:
        log_debug(ZONE,"unknown TCP command code 0x%04X",(unsigned)command);
        break;       
    }
}

void it_peer_send_message(contact c, imessage m)
{
    pool p = m->p;
    session s = c->s;
    unsigned len = sizeof(ICQTCP2MSG1) + sizeof(ICQTCP2MSG2) + sizeof(ICQTCP2TLN);
    unsigned payload_size;
    icqbyte *packet_base;
    icqbyte *payload;
    icqbyte *payload_end;
    PICQTCP2MSG1 pp1;
    PICQTCP2MSG2 pp2;
    PICQTCP2TLN ptl;

    log_debug(ZONE,"Sending peer message");

    payload_size = it_message_size(m);
    len += payload_size;
    packet_base = pmalloco(p,len);

    /* Get pointers to the beginning of the message packet and start formatting. */
    pp1 = (PICQTCP2MSG1)packet_base;
    payload = packet_base + sizeof(ICQTCP2MSG1);
    pp1->length = to_icqshort((unsigned short)(len - sizeof(icqshort)));
    pp1->uin2 = pp1->uin = to_icqlong(s->uin);
    pp1->version = to_icqshort(TCP_VERSION_NUMBER);
    pp1->command = to_icqshort(TCP_START);
    pp1->zero = 0;
    pp1->type = to_icqshort(m->type);
    pp1->textlen = to_icqshort((unsigned short)payload_size);
    payload_end = it_message_format(m,payload);
    pp2 = (PICQTCP2MSG2)payload_end;
    ptl = (PICQTCP2TLN)(payload_end + sizeof(ICQTCP2MSG2));
    pp2->ip = to_icqlong(s->addr_s.s_addr);
    pp2->real_ip = to_icqlong(s->addr.sin_addr.s_addr);
    pp2->port = to_icqshort(s->addr.sin_port);
    pp2->zero = 0;
    pp2->four = (icqbyte)4;
    pp2->status = to_icqshort(it_status_to_tcp(s->status));
    pp2->cmdtype = to_icqshort(it_format_tcp_msgcommand(TCP_MSG_REAL,s->status));
    ptl->tcpseq = to_icqlong(s->tcp_gen--);

    if (s->tcp_gen==0xFFFFFFFFUL)
        s->tcp_gen--;

    mio_write(c->m,NULL,packet_base,len);

    c->tcp_seq = from_icqlong(ptl->tcpseq);
    c->tcp_type = m->type;
}

void it_peer_ack_message(contact c, unsigned short type, unsigned long seq, char *response)
{
    session s = c->s;
    unsigned packet_size = sizeof(ICQTCP2MSG1) + sizeof(ICQTCP2MSG2) + sizeof(ICQTCP2TLN);
    unsigned payload_size; /* size of message payload */
    icqbyte *packet_base;  /* pointer to base packet */
    icqbyte *payload;      /* pointer to payload within packet */
    icqbyte *payload_end;  /* pointer to end of payload */
    PICQTCP2MSG1 pp1;      /* pointer to part 1 of packet */
    PICQTCP2MSG2 pp2;      /* pointer to part 2 of packet */
    PICQTCP2TLN ptl;       /* pointer to tail of packet */

    payload_size = j_strlen(response) + 1;
    packet_size += payload_size;
    packet_base = (icqbyte *)alloca(packet_size);

    /* Get pointers to the beginning of the message packet and start formatting. */
    pp1 = (PICQTCP2MSG1)packet_base;
    payload = packet_base + sizeof(ICQTCP2MSG1);
    pp1->length = to_icqshort((unsigned short)(packet_size - sizeof(icqshort)));
    pp1->uin2 = pp1->uin = to_icqlong(s->uin);
    pp1->version = to_icqshort(TCP_VERSION_NUMBER);
    pp1->command = to_icqshort(TCP_ACK);
    pp1->zero = 0;
    pp1->type = to_icqshort(type);
    pp1->textlen = to_icqshort((unsigned short)payload_size);

    if (response)
        memcpy(payload,response,payload_size);
    else
        *payload = '\0';

    payload_end = ++payload;

    pp2 = (PICQTCP2MSG2)payload_end;
    ptl = (PICQTCP2TLN)(payload_end + sizeof(ICQTCP2MSG2));

    pp2->ip = to_icqlong(s->addr_s.s_addr);
    pp2->real_ip = to_icqlong(s->addr.sin_addr.s_addr);
    pp2->port = to_icqshort(s->addr.sin_port);
    pp2->zero = 0;
    pp2->four = (icqbyte)4;

    pp2->status = to_icqshort(it_status_to_tcp(s->status));
    pp2->cmdtype = to_icqshort(it_format_tcp_msgcommand(TCP_MSG_AUTO,s->status));
    ptl->tcpseq = to_icqlong(seq);

#ifdef PACKET_DEBUG
    it_debug_dump(ZONE,packet_base,packet_size);
#endif

    mio_write(c->m,NULL,packet_base,packet_size);
}
