/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "icqtransport.h"

void it_presence(session s, jpacket jp)
{
    log_debug(ZONE,"Session[%s], handling presence",jid_full(s->id));

    switch (jpacket_subtype(jp))
    {
    case JPACKET__PROBE:
        if (jp->to->user)
        {
            contact c = it_contact_get(s,it_jid2uin(jp->to));
            if (c != NULL)
                it_contact_send_presence(c);
        }
        break;
    case JPACKET__AVAILABLE:
    {
        icqstatus status;
        char *text;

        if (jp->to->user)
            break;

        /* update users presence */
        s->p_db = ppdb_insert(s->p_db,jp->from,jp->x);

        text = xmlnode_get_tag_data(jp->x,"status");
        if (s->status_text)
        {
            if (text)
            {
                free(s->status_text);
                s->status_text = strdup(text);
            }
        }
        else if(text)
            s->status_text = strdup(text);

        status = it_show2status(xmlnode_get_tag_data(jp->x,"show"));
        if (status != s->status)
        {
            s->status = status;
            if (s->connected)
                it_server_send_status(s);
        }

        /* echo.... */
        jutil_tofrom(jp->x);
        xmlnode_hide(xmlnode_get_tag(jp->x,"status"));
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->x,"status"),
                             s->connected ? "Connected" : "Connecting...",-1);

        it_deliver(s->ti,jp->x);
        return;
    }

    case JPACKET__UNAVAILABLE:
        /* ignore offline presence */
        if (jp->to->user == NULL && xmlnode_get_tag(jp->x,"x") == NULL)
        {
            s->p_db = ppdb_insert(s->p_db,jp->from,jp->x);
            if (ppdb_primary(s->p_db,s->id) == NULL)
                it_session_end(s);
        }
    default:
        break;
    }

    xmlnode_free(jp->x);
}

void it_s10n(session s, jpacket jp)
{
	UIN_t uin;
	contact c;
    pool p = jp->p;

    if (jp->to->user == NULL)
    {
        if (jpacket_subtype(jp) == JPACKET__SUBSCRIBE)
        {
            jutil_tofrom(jp->x);
            xmlnode_put_attrib(jp->x,"type","subscribed");
            it_deliver(s->ti,jp->x);
        }
        else
            xmlnode_free(jp->x);
        return;
    }

    uin = it_jid2uin(jp->to);
	if (uin == 0)
	{
        jutil_error(jp->x,TERROR_BAD);
        it_deliver(s->ti,jp->x);
        return;
	}

    c = it_contact_get(s,uin);

	switch (jpacket_subtype(jp))
	{
    case JPACKET__SUBSCRIBE:
        log_debug(ZONE,"received JPACKET__SUBSCRIBE for %lu",uin);

        if (s->connected == 0)
        {
            jpq jq;
            log_debug(ZONE,"Queued");
            jq = pmalloco(p,sizeof(_jpq));
            jq->jp = jp;
            pth_msgport_put(s->queue,(pth_message_t *)jq);
            return;
        }

        if (c == NULL)
            c = it_contact_add(s,uin);

        jutil_tofrom(jp->x);
        xmlnode_put_attrib(jp->x,"type","subscribed");
        it_deliver(s->ti,jp->x);

        if (c->status == STATUS_NOT_IN_LIST)
            mtq_send(s->q,c->p,it_contact_subscribe,(void *) c);
        else
            it_contact_send_presence(c);
        return;

    case JPACKET__SUBSCRIBED:
        log_debug(ZONE,"received JPACKET__SUBSCRIBED for UIN#%lu",uin);

        if (c && c->asking)
        {
            c->asking = 0;
            it_message_send(c,it_message_create(jp->x,c->uin,MSGTYPE_AUTHORIZE));
        }
        else
            xmlnode_free(jp->x);

        break;

    case JPACKET__UNSUBSCRIBE:
        if (c && c->status != STATUS_NOT_IN_LIST)
        {
            if (s->connected == 0)
            {
                jpq jq;
                log_debug(ZONE,"Queued");
                jq = pmalloco(p,sizeof(_jpq));
                jq->jp = jp;
                pth_msgport_put(s->queue,(pth_message_t *)jq);
                return;
            }

            mtq_send(s->q,c->p,it_contact_unsubscribe,(void *) c);
        }

        xmlnode_free(jp->x);
        break;

    case JPACKET__UNSUBSCRIBED:
        log_debug(ZONE,"received JPACKET__UNSUBSCRIBED for UIN#%lu",uin);

        if (c && c->asking)
        {
            imessage m;

            m = it_message_create(jp->x,c->uin,MSGTYPE_AUTH_DENY,xmlnode_get_tag_data(jp->x,"status"));
            if (m)
                it_message_send(c,m);
            else
                xmlnode_free(jp->x);

            if (c->status == STATUS_NOT_IN_LIST && c->subscribing == 0)
                it_contact_remove(c);
        }
        else
            xmlnode_free(jp->x);
        break;
	}
}
