/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */

#define UDP_MAX_MSG_LEN       450                      /* Maximum length of a UDP message in chars */
#define UDP_MAX_DATA          (UDP_MAX_MSG_LEN + 9)    /* maximum data in a UDP packet */
#define UDP_MAX_CONTACTS      ((UDP_MAX_DATA - 1) / 4) /* max # of contacts in a UDP packet */

typedef unsigned char icqbyte;     /* single-byte value */
typedef unsigned short icqshort;   /* 2-byte value transferred via ICQ */
typedef unsigned long icqlong;     /* 4-byte value transferred via ICQ */
typedef unsigned long UIN_t;

struct icqheaderc_st
{
    icqshort version;
    icqlong zero;
    icqlong uin;
    icqlong sid;
    icqshort command;
    icqshort seq1;
    icqshort seq2;
    icqlong checkvalue;
} __attribute__ ((packed));

typedef struct icqheaderc_st ICQHDR5C, *PICQHDR5C;

struct tagICQBUFFER5C           /* buffer for an ICQ v5 client-to-server packet */
{
    ICQHDR5C hdr;                 /* packet header */
    icqbyte data[UDP_MAX_DATA];   /* packet data */
    icqbyte padding[4];           /* to prevent packet encryption from chomping memory */
} __attribute__ ((packed));
typedef struct tagICQBUFFER5C ICQBUFFER5C, *PICQBUFFER5C;

struct tagICQHDR5S              /* header for an ICQ v5 packet, server to client */
{
    icqshort version;             /* protocol version number (ICQ_VER_5) */
    icqbyte zero;                 /* always zero */
    icqlong sid;                  /* session ID - stays the same throughout a session */
    icqshort command;             /* command code */
    icqshort seq1;                /* first sequence number (starts at random value) */
    icqshort seq2;                /* second sequence number (starts at 1) */
    icqlong uin;                  /* UIN of client user */
    icqlong checkvalue;           /* header checkvalue */
    /* followed by packet data (variable) */
} __attribute__ ((packed));
typedef struct tagICQHDR5S ICQHDR5S, *PICQHDR5S;

typedef struct tagICQDECODE     /* buffer used to decode an incoming ICQ packet */
{
    unsigned short version;       /* version number of packet */
    unsigned short command;       /* command code */
    unsigned short seq1;          /* first sequence number */
    unsigned short seq2;          /* second sequence number */
    UIN_t uin;                    /* UIN of client user */
    icqbyte *data;                /* pointer to packet data */
    unsigned datalen;             /* length of packet data */
} ICQDECODE, *PICQDECODE;
typedef const ICQDECODE *PCICQDECODE;

typedef struct tagICQMETADECODE /* buffer used to decode an incoming SRV_META_USER packet */
{
    PCICQDECODE base_info;        /* pointer to the ICQDECODE buffer */
    unsigned short meta_command;  /* meta user command */
    icqbyte *data;                /* pointer to start of meta data */
    unsigned datalen;             /* length of meta data */
    unsigned success : 1;         /* was this a "success" packet? */
} ICQMETADECODE, *PICQMETADECODE;
typedef const ICQMETADECODE *PCICQMETADECODE;

struct tagICQTCP2INIT           /* ICQ TCP v2 initialization */
{
    icqshort length;              /* length of TCP packet following these bytes */
    icqbyte code;                 /* always TCP_INIT_PACKET_CODE */
    icqshort version;             /* protocol version number (TCP_VERSION_NUMBER) */
    icqshort zero1;               /* always 0 */
    icqshort port;                /* port number we listen on for messages */
    icqshort zero2;               /* always 0 */
    icqlong uin;                  /* UIN of other user */
    icqlong ip;                   /* IP address (LAN External) */
    icqlong real_ip;              /* "real" IP address (LAN Internal) */
    icqbyte four;                 /* always 4 */
    icqshort port2;               /* listening port for chat/file transfer */
    icqshort zero3;               /* always 0 */
} __attribute__ ((packed));
typedef struct tagICQTCP2INIT ICQTCP2INIT, *PICQTCP2INIT;

struct tagICQTCP2MSG1           /* part 1 of a TCP message packet */
{
    icqshort length;              /* length of TCP packet following these bytes */
    icqlong uin;                  /* sending UIN */
    icqshort version;             /* protocol version number (TCP_VERSION_NUMBER) */
    icqshort command;             /* command code */
    icqshort zero;                /* always 0 */
    icqlong uin2;                 /* sending UIN (again) */
    icqshort type;                /* message type */
    icqshort textlen;             /* message text length */
    /* followed by message text, and then ICQTCP2MSG2 */
} __attribute__ ((packed));
typedef struct tagICQTCP2MSG1 ICQTCP2MSG1, *PICQTCP2MSG1;

struct tagICQTCP2MSG2           /* part 2 of a TCP message packet */
{
    icqlong ip;                   /* sender's IP address (LAN External) */
    icqlong real_ip;              /* sender's "real" IP address (LAN Internal) */
    icqshort port;                /* port number of sender */
    icqshort zero;                /* always 0 */
    icqbyte four;                 /* always 4 */
    icqshort status;              /* current TCP status word */
    icqshort cmdtype;             /* message command type */
    /* followed by one of the "tail" structures */
} __attribute__ ((packed));
typedef struct tagICQTCP2MSG2 ICQTCP2MSG2, *PICQTCP2MSG2;

struct tagICQTCP2TLN            /* tail of a TCP message packet, normal */
{
    icqlong tcpseq;               /* TCP sequence number */
} __attribute__ ((packed));
typedef struct tagICQTCP2TLN ICQTCP2TLN, *PICQTCP2TLN;

/* ICQ version numbers */
#define ICQ_VER_5             0x0005

/* Checkvalue and encryption constants */
#define CHECKVALUE_INVERT     0x00FF00FFUL
#define CRYPTO_KEY_MAGIC_V5   0x68656C6CUL  /* magic constant for V5 header encryption */

/* Client-to-server command codes */
#define CMD_ACK               0x000A
#define CMD_SENDM             0x010E
#define CMD_LOGIN             0x03E8
#define CMD_CONTACT_LIST      0x0406
#define CMD_SEARCH_UIN        0x041A
#define CMD_SEARCH_USER       0x0424
#define CMD_KEEP_ALIVE        0x042E
#define CMD_KEEP_ALIVE2       0x051E
#define CMD_SEND_TEXT_CODE    0x0438
#define CMD_LOGIN_1           0x044C
#define CMD_INFO_REQ          0x0460
#define CMD_EXT_INFO_REQ      0x046A
#define CMD_CHANGE_PW         0x049C
#define CMD_STATUS_CHANGE     0x04d8
#define CMD_LOGIN_2           0x0528
#define CMD_UPDATE_INFO       0x050A
#define CMD_UPDATE_EXT_INFO   0x04B0
#define CMD_ADD_TO_LIST       0x053C
#define CMD_REQ_ADD_LIST      0x0456
#define CMD_QUERY_SERVERS     0x04BA
#define CMD_QUERY_ADDONS      0x04C4
#define CMD_NEW_USER_1        0x04EC
#define CMD_NEW_USER_REG      0x03FC
#define CMD_NEW_USER_INFO     0x04A6
#define CMD_ACK_MESSAGES      0x0442
#define CMD_MSG_TO_NEW_USER   0x0456
#define CMD_REG_NEW_USER      0x03FC
#define CMD_VIS_LIST          0x06AE
#define CMD_INVIS_LIST        0x06AE
#define CMD_TCP_REQUEST       0x015E
#define CMD_AUTH_UPDATE       0x0514
#define CMD_META_USER         0x064A

/* Server-to-client commands */
#define SRV_ACK               0x000A
#define SRV_LOGIN_REPLY       0x005A 
#define SRV_USER_ONLINE       0x006E
#define SRV_USER_OFFLINE      0x0078
#define SRV_USER_FOUND        0x008C
#define SRV_RECV_MESSAGE      0x00DC
#define SRV_END_OF_SEARCH     0x00A0
#define SRV_INFO_REPLY        0x0118
#define SRV_EXT_INFO_REPLY    0x0122
#define SRV_STATUS_UPDATE     0x01A4
#define SRV_X1                0x021C
#define SRV_X2                0x00E6
#define SRV_UPDATE            0x01E0
#define SRV_UPDATE_FAIL       0x01EA
#define SRV_UPDATE_EXT        0x00C8
#define SRV_NEW_UIN           0x0046
#define SRV_NEW_USER          0x00B4
#define SRV_QUERY             0x0082
#define SRV_SYSTEM_MESSAGE    0x01C2
#define SRV_MESSAGE           0x0104 
#define SRV_GO_AWAY           0x00F0
#define SRV_TRY_AGAIN         0x00FA  
#define SRV_FORCE_DISCONNECT  0x0028
#define SRV_MULTI_PACKET      0x0212
#define SRV_BAD_PASS          0x0064
#define SRV_META_USER         0x03DE
#define SRV_X_UNKNOWN_1       0x0050
#define SRV_TCP_REQUEST       0x015E

/* Magic numbers required for login */
#define LOGIN_MAGIC1          0x000000D5UL
#define LOGIN_MAGIC2          0x0000
#define LOGIN_MAGIC3          0x00000000UL
#define LOGIN_MAGIC4          0x00D50008UL
#define LOGIN_MAGIC5          0x00000050UL
#define LOGIN_MAGIC6          0x00000003UL
#define LOGIN_MAGIC7          0x00000000UL

/* Flags required for flags byte in login request */
#define LOGIN_FLAG1_FIREWALL  0x01    /* using firewall with dynamic TCP port */
#define LOGIN_FLAG1_PROXY     0x02    /* using SOCKS proxy */
#define LOGIN_FLAG1_USETCP    0x04    /* enable TCP communication */

/* Bit fields in server status value. */
#define SBIT_AWAY             0x00000001UL      /* away */
#define SBIT_DND              0x00000002UL      /* DND (usually w/occupied and away) */
#define SBIT_NA               0x00000004UL      /* NA (usually w/away) */
#define SBIT_OCCUPIED         0x00000010UL      /* occupied (usually w/away) */
#define SBIT_FREE_CHAT        0x00000020UL      /* free for chat */
#define SBIT_INVISIBLE        0x00000100UL      /* invisible */
#define SBIT_WEBAWARE         0x00010000UL      /* online status can be seen from Web */
#define SBIT_SHOW_IP          0x00020000UL      /* your IP address can be seen */

/* Magic numbers required for new account registration. */
#define REG_NEW_MAGIC1        0x000000A0UL
#define REG_NEW_MAGIC2        0x00002461UL
#define REG_NEW_MAGIC3        0x00A00000UL
#define REG_NEW_MAGIC4        0x00000000UL

#define TEXT_CODE_MAGIC       5

/* Instant message types from the server or TCP */
#define MSGTYPE_CONTACT_ADD   12
#define MSGTYPE_AUTH_REQUEST  6
#define MSGTYPE_AUTHORIZE     8
#define MSGTYPE_AUTH_DENY     7
#define MSGTYPE_URL           4
#define MSGTYPE_CONTACTS      19
#define MSGTYPE_TEXT          1
#define MSGTYPE_WEB_PAGER     13
#define MSGTYPE_EMAIL_PAGER   14
#define MSGTYPE_GREETING_CARD 26

/* Instant message types from TCP only */
#define MSGTYPE_CHAT          2
#define MSGTYPE_FILE          3
#define MSGTYPE_READ_AWAY     1000
#define MSGTYPE_READ_OCC      1001
#define MSGTYPE_READ_NA       1002
#define MSGTYPE_READ_DND      1003
#define MSGTYPE_READ_FFC      1004

/* Instant message type flags */
#define MSGTYPE_FLAG_MASS     0x0800

/* Message magic numbers */
#define MSG_AUTHORIZE_MAGIC   3
#define SEPARATOR_CHAR        '\xFE'  /* multi-string messages use this as a separator */

/* Meta user info commands */
#define META_GET_USER_INFO    0x04B0  /* retrieve info for a specified UIN */
#define META_SET_USER_INFO_2  0x03FC  /* set my "additional" user info (gender, homepage, etc.) */
#define META_SET_USER_INFO    0x03E8  /* set my user info (nick, first, last, email...) */
#define META_SET_USER_ABOUT   0x0406  /* set user "about" string */
#define META_SEARCH_UIN       0x051E  /* look for a user by UIN */
#define META_SET_PASSWORD     0x042E  /* set my password */
#define META_SET_SECURITY     0x0424  /* set my security settings */
#define META_SEARCH_EMAIL     0x0528  /* search for user by email address */
#define META_SEARCH_NAME      0x0514  /* search for user by name */
#define META_GET_USER_INFO_2  0x04CE  /* retrieve more info for a specified UIN? */
#define META_SET_USER_WORK    0x03F2  /* set my work address/contact info */
#define META_SET_INTERESTS    0x10B7  /* set my interests */
#define META_SET_AFFILIATIONs 0x4BCB  /* set my affiliations */

/* Meta user info server responses */
#define META_USER_INFO        0x00C8  /* server sends user info */
#define META_USER_INFO_2      0x00DC  /* server sends additional user info */
#define META_USER_PICTURE     0x010E  /* server sends user's picture? */
#define META_USER_WORK        0x00D2  /* server sends user's work address/contact info */
#define META_USER_ABOUT       0x00E6  /* server sends user's "about" string */
#define META_USER_X1          0x00F0  /* (unknown) */
#define META_USER_INFO_EOF    0x00FA  /* user information for this user is done */
#define META_USER_FOUND       0x0190  /* found a user - this gives the info */
#define META_LAST_USER_FOUND  0x019A  /* found a user - this gives info, is the last one */
#define META_ACK_SET_WORK     0x006E  /* ACK packet for META_SET_USER_WORK */
#define META_ACK_SET_MORE     0x0078  /* ACK packet for META_SET_USER_INFO_2 */
#define META_ACK_SET_INT      0x008C  /* ACK packet for META_SET_INTERESTS */
#define META_ACK_SET_AFFIL    0x0096  /* ACK packet for META_SET_AFFILIATIONS */
#define META_ACK_SET_PASSWORD 0x00AA  /* ACK packet for META_SET_PASSWORD */
#define META_ACK_SET_INFO     0x0064  /* ACK packet for META_SET_USER_INFO */
#define META_ACK_SET_SECURITY 0x000A  /* ACK packet for META_SET_SECURITY */
#define META_ACK_SET_ABOUT    0x0082  /* ACK packet for META_SET_USER_ABOUT */

/* TCP networking constants */
#define TCP_INIT_PACKET_CODE  0xFF    /* "code" value for init packet */
#define TCP_VERSION_NUMBER    0x0002  /* version number of peer-peer protocol */

/* TCP message command codes */
#define TCP_START             0x07EE  /* send message */
#define TCP_CANCEL            0x07D0  /* cancel message we just sent */
#define TCP_ACK               0x07DA  /* acknowledge the message */

/* TCP message command types */
#define TCP_MSG_AUTO          0x0000  /* automatic response */
#define TCP_MSG_REAL          0x0010  /* real message */
#define TCP_MSG_LIST          0x0020  /* message sent to the contact list */
#define TCP_MSG_URGENT        0x0040  /* urgent message */

/* TCP message command flags */
#define TCP_MSGF_S_INVISIBLE  0x0080  /* sending user is invisible */
#define TCP_MSGF_S_AWAY       0x0100  /* sending user is away */
#define TCP_MSGF_S_OCCUPIED   0x0200  /* sending user is occupied */
#define TCP_MSGF_S_NA         0x0800  /* sending user is N/A */
#define TCP_MSGF_S_DND        0x1000  /* sending user is DND */

/* TCP status values */
#define TCP_STAT_ONLINE       0x0000  /* online */
#define TCP_STAT_REFUSE       0x0001  /* refusing messages */
#define TCP_STAT_AWAY         0x0004  /* away */
#define TCP_STAT_OCCUPIED     0x0009  /* occupied */
#define TCP_STAT_DND          0x000A  /* DND */
#define TCP_STAT_NA           0x000E  /* N/A */
