/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "icqtransport.h"

void it_server_multi(session s,icqbyte *data);
void it_server_login_reply(session s);
void it_server_offline(session s, PCICQDECODE packet);
void it_server_reverse_tcp(session s, icqbyte *data);

void it_server_process_packet(session s, PCICQDECODE packet)
{
    log_debug(ZONE,"Session[%s] received ICQ packet",jid_full(s->id));

    switch (packet->command)
    {
    case SRV_MULTI_PACKET:   /* multi-packet - decode it */
        log_debug(ZONE,"SRV_MULTI_PACKET");
        it_server_multi(s,packet->data);
        break;

    case SRV_NEW_UIN:       /* new UIN received from the server */
        log_debug(ZONE,"SRV_NEW_UIN");
        s->uin = packet->uin;
        it_server_login(s);
        break;

    case SRV_LOGIN_REPLY:
        log_debug(ZONE,"SRV_LOGIN_REPLY");
        /* Get the IP address the server thinks we have. */
        s->addr_s.s_addr = get_icqlong(packet->data,12);
        it_server_login_reply(s);
        break;

    case SRV_X1:            /* server has acknowledged our contact lists */
        log_debug(ZONE,"SRV_X1");
        if (s->type == stype_normal)
        {
            jpq jq;
            s->connected = 1;
            while ((jq = (jpq) pth_msgport_get(s->queue)) != NULL)
                it_session_jpacket(s,jq->jp);
        }
        else
            mtq_send(s->q,s->p,it_session_register,(void *) s);

        break;

    case SRV_X2:            /* server has finished sending us offline messages */
        log_debug(ZONE,"SRV_X2");
        it_server_ack_messages(s);
        break;

    case SRV_RECV_MESSAGE:  /* server has sent us an "offline" message */
        log_debug(ZONE,"SRV_RECV_MESSAGE");
        it_server_offline(s,packet);
        break;

    case SRV_MESSAGE:
    {
        UIN_t sender;
        unsigned short msgtype, msglen;

        log_debug(ZONE,"SRV_MESSAGE");

        sender = get_icqlong(packet->data,0);
        msgtype = get_icqshort(packet->data,4);
        msglen = get_icqshort(packet->data,6);

        it_icq_message(s,sender,NULL,msgtype,packet->data+8,msglen);
        break;
    }

    case SRV_USER_OFFLINE:
    {
        contact c;

        log_debug(ZONE,"SRV_USER_OFFLINE");

        if ((c = it_contact_get(s,(UIN_t) get_icqlong(packet->data,0))) != NULL)
            it_contact_set_status(c,STATUS_OFFLINE);

        break;
    }

    case SRV_USER_ONLINE:
    {
        contact c;

        log_debug(ZONE,"SRV_USER_ONLINE");

        if ((c = it_contact_get(s,get_icqlong(packet->data,0))) != NULL)
        {
            int force_server = 1;  /* user's TCP communication flag */

            if (packet->data[16] & LOGIN_FLAG1_USETCP)
                force_server = ((packet->data[16] & LOGIN_FLAG1_PROXY) != 0);

            it_contact_update_addresses(c,get_icqlong(packet->data,4),get_icqshort(packet->data,8),
                                        get_icqlong(packet->data,12),force_server);

            it_contact_set_status(c,it_status_from_bits(get_icqlong(packet->data,17)));
        }
        break;
    }

    case SRV_STATUS_UPDATE:
    {
        contact c;

        log_debug(ZONE,"SRV_STATUS_UPDATE");

        if ((c = it_contact_get(s, get_icqlong(packet->data,0))) != NULL)
            it_contact_set_status(c,it_status_from_bits(get_icqlong(packet->data,4)));

        break;
    }

    case SRV_META_USER:
        log_debug(ZONE,"SRV_META_USER");
        it_meta_packet(s,packet);
        break;

    case SRV_TCP_REQUEST:
        log_debug(ZONE,"SRV_TCP_REQUEST");
        if (s->ti->disable_tcp == 0)
            it_server_reverse_tcp(s,packet->data);
        break;

    case SRV_INFO_REPLY:    /* info reply from server */
        log_debug(ZONE,"SRV_INFO_REPLY:");
        break;

    case SRV_EXT_INFO_REPLY: /* extended info reply from server */
        log_debug(ZONE,"SRV_EXT_INFO_REPLY");
        break;

    case SRV_USER_FOUND:
        log_debug(ZONE,"SRV_USER_FOUND");
        break;

    case SRV_END_OF_SEARCH:
        log_debug(ZONE,"SRV_END_OF_SEARCH");
        break;

    case SRV_GO_AWAY:
        log_debug(ZONE,"SRV_GO_AWAY");
        if (s->type != stype_normal)
            it_session_register_fail(s,(terror){502,"Remote Server Error, Please try again"});
        s->connected = 0;
        it_session_end(s);
        break;

    case SRV_FORCE_DISCONNECT:
        log_debug(ZONE,"SRV_FORCE_DISCONNECT");
        s->connected = 0;
        it_session_end(s);
        break;

    case SRV_BAD_PASS:
        log_debug(ZONE,"SRV_BAD_PASS");
        if (s->type != stype_normal)
            it_session_register_fail(s,(terror){401,"Bad password"});
        s->connected = 0;
        it_session_end(s);
        break;

    case SRV_TRY_AGAIN:
        log_debug(ZONE,"SRV_TRY_AGAIN");
        if (s->type != stype_normal)
            it_session_register_fail(s,(terror){502,"The ICQ server is temporarily unavailable, Please try again"});
        s->connected = 0;
        it_session_end(s);
        break;

    case 0x7108:  /* SRV_GO_TO_HELL */
    case 0x6400:  /* SRV_WHAT_THE_HELL? */
        log_debug(ZONE,"SRV_GO_TO_HELL");
        if (s->type != stype_normal)
            it_session_register_fail(s,(terror){502,"The ICQ server is temporarily unavailable, Please try again"});
        s->connected = 0;
        it_session_end(s);
	    break;

    default:
        log_debug(ZONE,"(unknown command data) %04",packet->command);
        break;
    }
}

/* breaks a SRV_MULTI_PACKET packet into its individual subpackets and processes them */
void it_server_multi(session s, icqbyte *data)
{
    ICQDECODE packet;
    PICQHDR5S ph5;
    
    unsigned count, raw_len;

    count = (unsigned)(*data++);  /* get number of packets */
    while (count > 0)
    {
        raw_len = (unsigned)get_icqshort(data,0);
        packet.version = (int)get_icqshort(data,2);
        ph5 = (PICQHDR5S)(data + 2);

        packet.command = from_icqshort(ph5->command);
        packet.seq1 = from_icqshort(ph5->seq1);
        packet.seq2 = from_icqshort(ph5->seq2);
        packet.uin = from_icqlong(ph5->uin);
        packet.data = data + (2 + sizeof(ICQHDR5S));
        packet.datalen = raw_len - sizeof(ICQHDR5S);

        if (packet.command == SRV_ACK)
            it_server_acked_packet(s,packet.seq1);
        else
            it_server_process_packet(s,&packet);

        data += (raw_len + 2);
        count--;
    }
}

void it_server_req_meta_info(session s, info_meta info, void *arg)
{
    log_debug(ZONE,"Session[%s], received login meta info",jid_full(s->id));

    if (info)
    {
        s->nick = pstrdup(s->p,info->nick);
        s->first = pstrdup(s->p,info->first);
        s->last = pstrdup(s->p,info->last);
        s->email = pstrdup(s->p,info->email);
    }
}

void it_server_login_reply(session s)
{
    log_debug(ZONE,"Session[%s], sending login reply",jid_full(s->id));

    /* start the keepalive for this session */
    s->timer = time(NULL);

    if (s->type != stype_register_new)
    {
        /* send the "finishing login" packet to the server */
        it_server_login_1(s);

        /* Send the user's contact list to the server. */
        it_server_send_contact_list(s);

        it_server_send_status(s);

        /* Request my account basic information from the server. */
        s->pend_info = it_meta_new(s,it_server_req_meta_info,NULL);
        it_server_request_meta_info(s,s->uin);
    }
    else
        it_server_send_info(s);
}

void it_server_offline(session s, PCICQDECODE packet)
{
    UIN_t sender;
    char timestamp[18];
    unsigned short msgtype, msglen;
    int year, mon, mday, hour, min;
    int ret;

    sender = get_icqlong(packet->data,0);
    msgtype = get_icqshort(packet->data,10);
    msglen = get_icqshort(packet->data,12);

    year = (int) get_icqshort(packet->data,4);
    mon = (int) (packet->data[6]);
    mday = (int) (packet->data[7]);
    hour = (int) (packet->data[8]);
    min = (int) (packet->data[9]);

    ret = snprintf(timestamp, 18, "%d%02d%02dT%02d:%02d:%02d", year,mon,mday,hour,min,0);

    it_icq_message(s,sender,ret != -1 ? timestamp : NULL,msgtype,packet->data + 14,msglen);
}

void it_server_reverse_tcp(session s, icqbyte *data)
{
    contact c;

    c = it_contact_get(s,get_icqlong(data,0));
    if (c)
    {
        struct sockaddr_in addr;

        memset(&addr,0,sizeof(struct sockaddr_in));
        addr.sin_family = AF_INET;
        addr.sin_addr.s_addr = get_icqlong(data,4);
        addr.sin_port = htons(get_icqshort(data,8));

        memcpy(&(c->addr),&addr,sizeof(struct sockaddr_in));
        c->force_server = 0;
        it_peer_connect(c);
    }
}
