/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "icqtransport.h"

session it_session_create(iti ti, jpacket jp)
{
    pool p;
    session s;
    jid id = jp->from;

    log_debug(ZONE,"Creating session for %s",jid_full(id));

    p = pool_new();
    s = pmalloco(p,sizeof(_session));
    s->p = p;
    s->ti = ti;
    s->q = mtq_new(p);

    s->id = jid_new(p,jid_full(jp->from));
    jid_set(s->id,NULL,JID_RESOURCE);
    s->from = jid_new(p,jid_full(jp->to));
    jid_set(s->from,"registered",JID_RESOURCE);

    s->timer = time(NULL);

    s->queue = pth_msgport_create("it_session_queue");
    s->rsend = -1;
    s->status = STATUS_OFFLINE;
    s->type = stype_normal;
    s->addr.sin_family = AF_INET;
    s->addr.sin_addr.s_addr = 0x0100007FUL;  /* 127.0.0.1 in network byte order */

    s->sid = (unsigned long) random();
    s->seq1 = (unsigned short)(random() & 0x7FFF);
    s->seq2 = 1;

    ghash_put(ti->sessions,jid_full(s->id),(void *) s);

    s->next = ti->session_list;
    ti->session_list = s;
    if (s->next) s->next->prev = s;
    ++ti->sessions_count;

    return s;
}

void it_session_start(void *arg)
{
    session s = (session) arg;

    log_debug(ZONE,"Session[%s], starting",jid_full(s->id));

    if (s->type == stype_normal)
        it_contact_load_roster(s);

    if (it_server_connect(s))
    {
        log_debug(ZONE,"Failed to connect to ICQ server!");
        it_session_end(s);
        return;
    }

    if (s->ti->disable_tcp == 0 && it_peer_listen(s))
        log_debug(ZONE,"Session[%s], failed to create peer socket!",jid_full(s->id));

    if (s->uin == 0)
        it_server_new_UIN(s);
    else
        it_server_login(s);
}

void it_session_jpacket(session s, jpacket jp)
{
    switch (jp->type)
    {
    case JPACKET_IQ:
        it_iq(s,jp);
        break;
    case JPACKET_S10N:
        it_s10n(s,jp);
        break;
    case JPACKET_MESSAGE:
        it_message(s,jp);
        break;
    case JPACKET_PRESENCE:
        it_presence(s,jp);
        break;
    default:
        xmlnode_free(jp->x);
    }
}

void it_session_register_fail(session s, terror e)
{
    jpq jq;
    xmlnode x;

    jq = (jpq) pth_msgport_get(s->queue);
    x = jq->jp->x;
    s->type = stype_normal;

    jutil_error(x,e);
    it_deliver(s->ti,x);
}

void it_session_register(void *arg)
{
    session s = (session) arg;
    iti ti = s->ti;
    jpacket jp;
    xmlnode xreg;
    pool p;
    jpq jq;
    char uin[16];

    jq = (jpq) pth_msgport_get(s->queue);
    jp = jq->jp;
    p = jp->p;
    s->type = stype_normal;

    xmlnode_hide(xmlnode_get_tag(jp->iq,"instructions"));
    xreg = xmlnode_dup(jp->iq);
    xmlnode_hide(xmlnode_get_tag(xreg,"username"));
    snprintf(uin,16,"%lu",s->uin);
    xmlnode_insert_cdata(xmlnode_insert_tag(xreg,"username"),uin,-1);

    if (xdb_set(ti->xc,it_xdb_id(p,jp->from,s->from->server),NS_REGISTER,xreg) == 0)
    {
        xmlnode x;
        char *from;

        from = jid_full(s->from);

        x = jutil_presnew(JPACKET__SUBSCRIBE,jid_full(s->id),NULL);
        xmlnode_put_attrib(x,"from",from);
        it_deliver(ti,x);

        x = jutil_presnew(JPACKET__PROBE,jid_full(s->id),NULL);
        xmlnode_put_attrib(x,"from",from);
        it_deliver(ti,x);

        jutil_iqresult(jp->x);
        it_deliver(ti,jp->x);

        s->connected = 1;
        while ((jq = (jpq) pth_msgport_get(s->queue)) != NULL)
            it_session_jpacket(s,jq->jp);
    }
    else
    {
        jutil_error(jp->x,TERROR_UNAVAIL);
        it_session_end(s);
        it_deliver(ti,jp->x);
    }
}

void _it_session_end(void *arg)
{
    session s = (session) arg;
    iti ti = s->ti;
    jpq jq;
    contact c;
    xmlnode x;

    log_debug(ZONE,"Session[%s], exiting",jid_full(s->id));

    if (s->server)
    {
        if (s->connected)
            it_server_text_code(s,"B_USER_DISCONNECTED");
        s->server->cb_arg = NULL;
        mio_close(s->server);
    }

    if (s->peer_accept)
        mio_close(s->peer_accept);

    ghash_remove(ti->sessions,jid_full(s->id));

    if (s->prev)
    {
        s->prev->next = s->next;

        if (s->next)
            s->next->prev = s->prev;
    }
    else
    {
        /* we are at the head */
        if (s->next)
        {
            ti->session_list = s->next;
            ti->session_list->prev = NULL;
        }
        else
            ti->session_list = NULL;
    }

    if (s->contacts)
    {
        x = jutil_presnew(JPACKET__UNAVAILABLE,jid_full(s->id),NULL);

        c = s->contacts;
        while (c)
        {
            pool p = c->p;
            xmlnode pres;

            pres = xmlnode_dup(x);
            xmlnode_put_attrib(pres,"from",jid_full(it_uin2jid(p,c->uin,s->from->server)));
            it_deliver(ti,pres);

            if (c->m)
                it_peer_close(c);

            while ((jq = (jpq) pth_msgport_get(c->queue)) != NULL)
            {
                xmlnode msg = jq->jp->x;
                jutil_error(msg,TERROR_NOTFOUND);
                it_deliver(ti,msg);
            }

            pth_msgport_destroy(c->queue);
            c = c->next;
            pool_free(p);
        }

        xmlnode_put_attrib(x,"from",jid_full(s->from));
        it_deliver(ti,x);
    }
    else if (s->type != stype_normal)
        it_session_register_fail(s,TERROR_NOTACCEPTABLE);

    while ((jq = (jpq) pth_msgport_get(s->queue)) != NULL)
    {
        xmlnode x = jq->jp->x;
        jutil_error(x,TERROR_NOTFOUND);
        it_deliver(ti,x);
    }
    pth_msgport_destroy(s->queue);

    if (s->pq_head)
    {
        pqueue cur = s->pq_head;
        pool p;

        while (cur)
        {
            p = cur->p;
            cur = cur->next;
            pool_free(p);
        }
    }

    if (s->status_text)
        free(s->status_text);

    ppdb_free(s->p_db);
    pool_free(s->p);

    --ti->sessions_count;
}

void it_session_end(session s)
{
    if (s->exit_flag == 1)
        return;

    log_debug(ZONE,"Killing session[%s]",jid_full(s->id));

    s->exit_flag = 1;

    mtq_send(s->q,s->p,_it_session_end,(void *) s);
}
