/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "icqtransport.h"
#include <sys/utsname.h>

void it_unknown_presence(void *arg);
void it_unknown_reg_get(iti ti, jpacket jp);
void it_unknown_reg_set(iti ti, jpacket jp);
void it_unknown_bounce(void *arg);

void it_unknown_iq(iti ti, jpacket jp)
{
    char *ns;

    if (jp->to->user != NULL)
    {
        jp->aux1 = (void *) ti;
        mtq_send(NULL,jp->p,it_unknown_bounce,(void *) jp);
        return;
    }

    ns = xmlnode_get_attrib(jp->iq,"xmlns");

    switch (jpacket_subtype(jp))
    {
    case JPACKET__SET:
        if (j_strcmp(ns,NS_REGISTER) == 0)
            it_unknown_reg_set(ti,jp);
        else
        {
            jutil_error(jp->x,TERROR_NOTALLOWED);
            it_deliver(ti,jp->x);
        }

        break;
    case JPACKET__GET:
        if (j_strcmp(ns,NS_REGISTER) == 0)
            it_unknown_reg_get(ti,jp);
        else if (j_strcmp(ns,NS_BROWSE) == 0)
            it_iq_browse_server(ti,jp);
        else if (j_strcmp(ns,NS_VERSION) == 0)
            it_iq_version(ti,jp);
        else if (j_strcmp(ns,NS_TIME) == 0)
            it_iq_time(ti,jp);
        else if (j_strcmp(ns,NS_VCARD) == 0)
            it_iq_vcard_server(ti,jp);
        else if (j_strcmp(ns,NS_LAST) == 0)
            jp->to->user == NULL ? it_iq_last(ti,jp) : xmlnode_free(jp->x);
        else
        {
            jutil_error(jp->x,TERROR_NOTALLOWED);
            it_deliver(ti,jp->x);
        }
        break;

    default:
        jutil_error(jp->x,TERROR_NOTALLOWED);
        it_deliver(ti,jp->x);
    }
}

void it_unknown(iti ti, jpacket jp)
{
    log_debug(ZONE,"Packet from unknown user '%s'",jid_full(jp->from));

    switch (jp->type)
    {
    case JPACKET_IQ:
        it_unknown_iq(ti,jp);
        break;

    case JPACKET_MESSAGE:
    case JPACKET_S10N:
        jp->aux1 = (void *) ti;
        mtq_send(NULL,jp->p,it_unknown_bounce,(void *) jp);
        break;

    case JPACKET_PRESENCE:
        if (jpacket_subtype(jp) == JPACKET__AVAILABLE && jp->to->user == NULL &&
            j_strncmp(jp->to->resource,"registered",10) == 0) {

            jp->aux1 = (void *) ti;
            mtq_send(NULL,xmlnode_pool(jp->x),it_unknown_presence,(void *) jp);
            break;
        }
    default:
        xmlnode_free(jp->x);
    }
}

void it_unknown_presence(void *arg)
{
    jpacket jp = (jpacket) arg;
    iti ti = jp->aux1;
    xmlnode reg;
    session s;

    reg = xdb_get(ti->xc,it_xdb_id(jp->p,jp->from,jp->to->server),NS_REGISTER);
    if (reg == NULL)
    {
        log_debug(ZONE,"Registration not found for %s",jid_full(jp->from));
        xmlnode_free(jp->x);
        return;
    }

    s = it_session_create(ti,jp);
    s->type = stype_normal;
    s->uin = it_strtouin_safe(xmlnode_get_tag_data(reg,"username"));
    s->passwd = convert_utf8_to_windows(s->p,xmlnode_get_tag_data(reg,"password"));

    if (s->uin == 0 || s->passwd == NULL)
    {
        log_warn(jp->to->server,"User %s has invalid registration setting",jid_full(jp->from));
        xmlnode_free(jp->x);
        return;
    }

    s->nick = convert_utf8_to_windows(s->p,xmlnode_get_tag_data(reg,"nick"));
    s->first = convert_utf8_to_windows(s->p,xmlnode_get_tag_data(reg,"first"));
    s->last = convert_utf8_to_windows(s->p,xmlnode_get_tag_data(reg,"last"));
    s->email = convert_utf8_to_windows(s->p,xmlnode_get_tag_data(reg,"email"));

    xmlnode_free(reg);

    /* update users presence */
    s->p_db = ppdb_insert(s->p_db,jp->from,jp->x);
    s->status = it_show2status(xmlnode_get_tag_data(jp->x,"show"));

    it_session_start((void *) s);

    /* Echo back the users presence */
    jutil_tofrom(jp->x);
    xmlnode_hide(xmlnode_get_tag(jp->x,"status"));
    xmlnode_insert_cdata(xmlnode_insert_tag(jp->x,"status"),"Connecting...",-1);
    it_deliver(ti,jp->x);
}

void it_unknown_reg_get(iti ti, jpacket jp)
{
    xmlnode q;

    jutil_iqresult(jp->x);

    q = xmlnode_insert_tag(jp->x,"query");
    xmlnode_put_attrib(q,"xmlns",NS_REGISTER);

    xmlnode_insert_tag(q,"username");
    xmlnode_insert_tag(q,"nick");
    xmlnode_insert_tag(q,"first");
    xmlnode_insert_tag(q,"last");
    xmlnode_insert_tag(q,"email");
    xmlnode_insert_tag(q,"password");

    xmlnode_insert_cdata(xmlnode_insert_tag(q,"instructions"),ti->reg,-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(q,"key"),jutil_regkey(NULL,jid_full(jp->from)),-1);

    it_deliver(ti,jp->x);
}

void it_unknown_reg_set(iti ti, jpacket jp)
{
    session s;
    xmlnode q = jp->iq;
    UIN_t uin;
    jpq jq;
    char *user, *pass, *key;

    key = xmlnode_get_tag_data(q,"key");
    if (key == NULL || !jutil_regkey(key,jid_full(jp->from)))
    {
        jutil_error(jp->x,(terror){406,"Invalid key"});
        it_deliver(ti,jp->x);
        return;
    }

    pass = xmlnode_get_tag_data(q,"password");
    user = xmlnode_get_tag_data(q,"username");
    uin = it_strtouin_safe(user);

    if (pass == NULL || (uin == 0 && user))
    {
        jutil_error(jp->x,TERROR_NOTACCEPTABLE);
        it_deliver(ti,jp->x);
        return;
    }

    s = it_session_create(ti,jp);
    s->type = uin ? stype_register : stype_register_new;
    s->uin = uin;
    s->passwd = convert_utf8_to_windows(s->p,pass);

    /* these aren't required */
    s->nick = convert_utf8_to_windows(s->p,xmlnode_get_tag_data(q,"nick"));
    s->first = convert_utf8_to_windows(s->p,xmlnode_get_tag_data(q,"first"));
    s->last = convert_utf8_to_windows(s->p,xmlnode_get_tag_data(q,"last"));
    s->email = convert_utf8_to_windows(s->p,xmlnode_get_tag_data(q,"email"));

    jq = pmalloco(jp->p,sizeof(_jpq));
    jq->jp = jp;
    pth_msgport_put(s->queue,(pth_message_t *)jq);

    mtq_send(s->q,s->p,it_session_start,(void *) s);
}

void it_unknown_bounce(void *arg)
{
    jpacket jp = (jpacket) arg;
    iti ti = (iti) jp->aux1;
    xmlnode reg;

    reg = xdb_get(ti->xc,it_xdb_id(jp->p,jp->from,jp->to->server),NS_REGISTER);
    if (reg != NULL)
    {
        jutil_error(jp->x,(terror){404,"Session Not Found"});
        xmlnode_free(reg);
    }
    else
        jutil_error(jp->x,TERROR_REGISTER);

    it_deliver(ti,jp->x);
}
