/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "icqtransport.h"
#include "tables.h"

int it_bigendian;

void it_check_endian(void)
{
    int i;
    char passwd[10];
   
    passwd[0] = 1;
    passwd[1] = 0;
    passwd[2] = 0;
    passwd[3] = 0;
    passwd[4] = 0;
    passwd[5] = 0;
    passwd[6] = 0;
    passwd[7] = 0;
    passwd[8] = 0;
    passwd[9] = 0;

    i = *(icqlong *) passwd;

    if (i == 1)
    {
        log_debug(ZONE,"Little endian");
        it_bigendian = 0;
    }
    else
    {
        log_debug(ZONE,"Big endian");
        it_bigendian = 1;
    }
}

jid it_xdb_id(pool p, jid id, char *server)
{
    char *buf;
    buf = spools(p,id->user,"%",id->server,"@",server,p);
    return jid_new(p,buf);
}

UIN_t it_strtouin_safe(char *uin)
{
    if (uin == NULL)
        return 0;

    return strtoul(uin,NULL,10);
}

jid it_uin2jid(pool p, UIN_t uin, char *server)
{
    jid rc;
    char buffer[16];

    rc = (jid) pmalloco(p,sizeof(struct jid_struct));

    rc->p = p;
    rc->server = pstrdup(p,server);
    snprintf(buffer,16,"%lu",uin);
    rc->user = pstrdup(p,buffer);

    return rc;
}

icqstatus it_show2status(const char *show)
{
    if (!show)
        return STATUS_ONLINE;

    if (j_strcmp(show,"away")==0)
        return STATUS_AWAY;
    if (j_strcmp(show,"busy")==0)
        return STATUS_OCCUPIED;
    if (j_strcmp(show,"chat")==0)
        return STATUS_FREE_CHAT;
    if (j_strcmp(show,"dnd")==0)
        return STATUS_DND;
    if (j_strcmp(show,"xa")==0)
        return STATUS_NA;

    return STATUS_ONLINE;
}

char *it_strrepl(pool p, const char *orig, const char *find, const char *replace)
{
    const char *loc;
    char *newstr, *temp, *tempstr;
    unsigned int olen, flen, rlen, i;

    if (!orig || !find || !replace || !p)
        return NULL;  /* oops */

    olen = strlen(orig);
    flen = strlen(find);
    rlen = strlen(replace);

    temp = strstr(orig,find);
    if (!temp)
        return (p ? pstrdup(p,orig) : j_strdup(orig));

    /* count up any instances of the string we can find */
    i = 0;
    while (temp)
    { /* advance past current instance and get a count */
        temp += flen;
        i++;
        temp = strstr(temp,find);
    }

    /* allocate a buffer for the new string */
    newstr = pmalloc(p,(olen + ((rlen - flen) * i) + 1) * sizeof(char));
    tempstr = newstr;
    loc = orig;

    while((temp = strstr(loc,find))!=NULL)
    { /* copy pieces into the new buffer */
        memcpy(tempstr,loc,(int)(temp-loc));
        tempstr += (int)(temp-loc);

        memcpy(tempstr,replace,rlen);
        tempstr += rlen;

        loc = temp + flen;
    }

    strcpy(tempstr,loc);  /* copy the last piece */

    return newstr;
}

/* calculates and returns the ICQ version 5 checkvalue for a packet */
unsigned long it_checkvalue5(const icqbyte *packet,unsigned packet_len)
{
    unsigned long check, check2;         /* checkvalue computations */
    unsigned int ofs;                    /* used in getting checkvalue */

    check = (((unsigned long)(packet[8])) << 24) | (((unsigned long)(packet[4])) << 16) |
        (((unsigned long)(packet[2])) << 8) | ((unsigned long)(packet[6]));

    ofs = sizeof(ICQHDR5C) + (random() % (packet_len - sizeof(ICQHDR5C)));
    check2 = (ofs << 8) + packet[ofs];
    ofs = random() % 0xFF;
    check2 = (check2 << 8) + ofs;
    check2 = (check2 << 8) + icq_V5_check_data[ofs];

    return check ^ (check2 ^ CHECKVALUE_INVERT);
}

/* encodes a packet according to the ICQ version 5 algorithm */
void it_scramble5(void *buffer, unsigned len, unsigned long checkvalue)
{
    unsigned i;                   /* loop counter */
    unsigned long ccs[5];         /* checkvalue scrambling buffer */

    /* Set up the encryption loop. */
    ccs[0] = (len * CRYPTO_KEY_MAGIC_V5) + checkvalue;  /* crypto key */
    for (i=0xA; i<(len+3); i+=4)
    {
        ccs[1] = get_icqlong(buffer,i);
        put_icqlong(buffer,i,ccs[1] ^ (ccs[0] + icq_V5_check_data[i & 0xFF]));
    }

    /* Scramble the checkvalue and insert it into the packet. */
    ccs[0] = (checkvalue & 0x0000001FUL) << 0x0C;
    ccs[1] = (checkvalue & 0x03E003E0UL) << 0x01;
    ccs[2] = (checkvalue & 0xF8000400UL) >> 0x0A;
    ccs[3] = (checkvalue & 0x0000F800UL) << 0x10;
    ccs[4] = (checkvalue & 0x041F0000UL) >> 0x0F;
    put_icqlong(buffer,0x14,ccs[0]+ccs[1]+ccs[2]+ccs[3]+ccs[4]);
}

/*  Transforms a 2-byte value from ICQ packet format into native format. */
inline unsigned short from_icqshort(icqshort arg)
{
    return it_bigendian ? (unsigned short)(((arg << 8) & 0xFF00) | ((arg >> 8) & 0x00FF)) : arg;
}

/* Transforms a 2-byte value from native format into ICQ packet format. */
inline icqshort to_icqshort(unsigned short arg)
{
    return it_bigendian ? (unsigned short)(((arg << 8) & 0xFF00) | ((arg >> 8) & 0x00FF)) : arg;
}

/* Transforms a 4-byte value from ICQ packet format into native format. */
inline unsigned long from_icqlong(icqlong arg)
{
    return it_bigendian ? (((arg << 24) & 0xFF000000UL) | ((arg << 8) & 0x00FF0000UL)
                           | ((arg >> 8) & 0x0000FF00UL)  | ((arg >> 24) & 0x000000FFUL)) : arg;
}

/* Transforms a 4-byte value from native format into ICQ packet format. */
inline icqlong to_icqlong(unsigned long arg)
{
    return it_bigendian ? (((arg << 24) & 0xFF000000UL) | ((arg << 8) & 0x00FF0000UL)
                           | ((arg >> 8) & 0x0000FF00UL)  | ((arg >> 24) & 0x000000FFUL)) : arg;
}

/* Retrieves a 2-byte value from some location in a buffer and converts it
   from ICQ packet format to native format. */
unsigned short get_icqshort(const void *buffer, unsigned offset)
{
    register const icqbyte *p = (const icqbyte *)buffer;
    p += offset;
    return from_icqshort(*((icqshort *)p));
}

/* Converts a 2-byte value from native format to ICQ packet format and stores
   it at some location in a buffer. */
void put_icqshort(void *buffer, unsigned offset, unsigned short value)
{
    register icqbyte *p = (icqbyte *)buffer;
    p += offset;
    *((icqshort *)p) = to_icqshort(value);
}

/* Retrieves a 4-byte value from some location in a buffer and converts it
   from ICQ packet format to native format. */
unsigned long get_icqlong(const void *buffer, unsigned offset)
{
    register const icqbyte *p = (const icqbyte *)buffer;
    p += offset;
    return from_icqlong(*((icqlong *)p));
}

/* Converts a 4-byte value from native format to ICQ packet format and stores
   it at some location in a buffer. */
void put_icqlong(void *buffer, unsigned offset, unsigned long value)
{
    register icqbyte *p = (icqbyte *)buffer;
    p += offset;
    *((icqlong *)p) = to_icqlong(value);
}

/* Converts the status bits returned from the server into an ICQ status
	value.  This must be done by testing individual bits in a specific order,
	as the precise pattern of bits can vary for individual states. */
icqstatus it_status_from_bits(unsigned long bits)
{
    if (bits & SBIT_DND)
        return STATUS_DND;
    if (bits & SBIT_OCCUPIED)
        return STATUS_OCCUPIED;
    if (bits & SBIT_NA)
        return STATUS_NA;
    if (bits & SBIT_AWAY)
        return STATUS_AWAY;
    if (bits & SBIT_FREE_CHAT)
        return STATUS_FREE_CHAT;
    if (bits & SBIT_INVISIBLE)
        return STATUS_INVISIBLE;

    return STATUS_ONLINE;
}

/* Converts an icqstatus value into a combination of status bits that gets sent
   to the server.  The mapping is based on that used by the Windows clients. */
unsigned long it_status_to_bits(icqstatus status)
{
    static const unsigned long s_table[] =
        { 0, SBIT_AWAY, SBIT_AWAY|SBIT_DND|SBIT_OCCUPIED, SBIT_NA|SBIT_AWAY,
          SBIT_OCCUPIED|SBIT_AWAY, SBIT_FREE_CHAT, SBIT_INVISIBLE };

    if (((int)status)<0)
        return 0xFFFFFFFFUL;
    else
        return s_table[status] | SBIT_WEBAWARE;
}

/* Converts an ICQ status value to a TCP status value. */
unsigned short it_status_to_tcp(icqstatus status)
{
    static const unsigned short s_conversion_table[] =
        {
            TCP_STAT_ONLINE,    /* = STATUS_ONLINE */
            TCP_STAT_AWAY,      /* = STATUS_AWAY */
            TCP_STAT_DND,       /* = STATUS_DND */
            TCP_STAT_NA,        /* = STATUS_NA */
            TCP_STAT_OCCUPIED,  /* = STATUS_OCCUPIED */
            TCP_STAT_ONLINE,    /* = STATUS_FREE_CHAT */
            TCP_STAT_REFUSE     /* = STATUS_INVISIBLE */
        };

    if (status<0)
        return TCP_STAT_ONLINE;
    else
        return s_conversion_table[status];
}

/* Adds "modifier flags" to a TCP message command constant based on a current status value. */
unsigned short it_format_tcp_msgcommand(unsigned short base_cmd, icqstatus status)
{
    static const unsigned short s_modifiers_table[] =
        {
            0,                    /* = STATUS_ONLINE */
            TCP_MSGF_S_AWAY,      /* = STATUS_AWAY */
            TCP_MSGF_S_DND,       /* = STATUS_DND */
            TCP_MSGF_S_NA,        /* = STATUS_NA */
            TCP_MSGF_S_OCCUPIED,  /* = STATUS_OCCUPIED */
            0,                    /* = STATUS_FREE_CHAT */
            TCP_MSGF_S_INVISIBLE  /* = STATUS_INVISIBLE */
        };

    if (status<0)
        return base_cmd;
    else
        return (base_cmd | s_modifiers_table[status]);
}

/* Converts a TCP status value to an ICQ status value. */
icqstatus it_status_from_tcp(unsigned short tcpstat)
{
    switch (tcpstat)
    {
    case TCP_STAT_ONLINE:
        return STATUS_ONLINE;

    case TCP_STAT_AWAY:
        return STATUS_AWAY;

    case TCP_STAT_NA:
        return STATUS_NA;

    case TCP_STAT_DND:
        return STATUS_DND;

    case TCP_STAT_OCCUPIED:
        return STATUS_OCCUPIED;

    case TCP_STAT_REFUSE:
        return STATUS_INVISIBLE;

    default:
        break;
    }

    return STATUS_OFFLINE;
}

/* Returns the country name corresponding to a given ICQ country code. */
const char *it_country_from_code(unsigned short code)
{
    register int i;  /* loop counter */
    for (i=0; s_countries[i].code!=0; i++)
        if (s_countries[i].code==code)
            return s_countries[i].fullname;

    return NULL;
}

/* decodes and verifies a version 5 server-to-client packet, returning an appropriate PACKET_STATUS code */
int it_test_v5_server(session s, void *buffer, unsigned len)
{
    PICQHDR5S hdr;        /* pointer to header */
    unsigned short tmps;  /* temporary short value */
    unsigned long tmpl;   /* temporary long value */

    /* Verify that the packet is long enough to be a v5 server packet. */
    if (len < sizeof(ICQHDR5S))
    {
        log_debug(ZONE,"test_v5_server: packet must be at least %u bytes long, but is only %u",sizeof(ICQHDR5S),len);
        return 0;
    }

    /* Verify that the version number is correct. */
    hdr = (PICQHDR5S)buffer;
    tmps = from_icqshort(hdr->version);
    if (tmps != ICQ_VER_5)
    {
        log_debug(ZONE,"test_v5_server: version should be 0x%04X, but is 0x%04X",ICQ_VER_5,tmps);
        return 0;
    }

#if 0   /* GnomeICU doesn't seem to need this */
    /* Verify that the session ID is correct. */
    tmpl = from_icqlong(hdr->sid);
    if (tmpl != s->sid)
    {
        log_debug(ZONE,"test_v5_server: session ID should be 0x%08lX, but is 0x%08lX",s->sid,tmpl);
        return 0;
    }
#endif

    /* Verify that the packet is for us. */
    tmpl = from_icqlong(hdr->uin);
    if ((tmpl != s->uin) && (from_icqshort(hdr->command) != SRV_NEW_UIN))
    {
        log_debug(ZONE,"test_v5_server: UIN should be %lu, but is %lu",s->uin,tmpl);
        return 0;
    }

    return 1;
}

void it_debug_dump(const char *zone, const void *buffer, unsigned nbytes)
{
    const unsigned char *dbuf = (const unsigned char *)buffer;  /* the actual buffer */
    unsigned base;    /* base index for a row */
    unsigned limit;   /* number of bytes to put in this row */
    register unsigned i;   /* loop counter */
    char hexbuf[56], ascbuf[32], tmp[8];  /* ASCII conversion buffers */

    if (!debug_flag)
        return;

    fprintf(stderr,"%s: [DUMP] (%u bytes)\n",zone,nbytes);
    for (base=0; base<nbytes; base+=16)
    { /* start by determining how many bytes go on this row */
        limit = nbytes - base;
        if (limit>16)
            limit = 16;

        /* clear buffers */
        hexbuf[0] = ascbuf[0] = '\0';

        for (i=0; i<limit; i++)
        {
            /* convert the hex value to ASCII and put it in */
            sprintf(tmp,"%02X ",(unsigned)(dbuf[base+i]));
            strcat(hexbuf,tmp);

            /* get the ascii character for this byte and add it */
            if (isprint(dbuf[base+i]))
                tmp[0] = (char)(dbuf[base+i]);
            else
                tmp[0] = '.';

            tmp[1] = '\0';
            strcat(ascbuf,tmp);

            if (i==7)
            { /* concatenate the midpoint stuff */
                strcat(hexbuf,"- ");
                strcat(ascbuf," ");
            }
        }

        /* output this row */
        fprintf(stderr,"%08X: %-50s | %s\n",base,hexbuf,ascbuf);
    }

    fprintf(stderr,"%s: [END DUMP]\n",zone);
}

void it_delay(xmlnode x, char *ts)
{
    xmlnode delay;

    delay = xmlnode_insert_tag(x,"x");
    xmlnode_put_attrib(delay,"xmlns",NS_DELAY);
    xmlnode_put_attrib(delay,"from",xmlnode_get_attrib(x,"to"));
    xmlnode_put_attrib(delay,"stamp",ts);
}
